import os
import shutil
import unittest
from medcat.vocab import Vocab


class CATTests(unittest.TestCase):

    def setUp(self) -> None:
        self.undertest = Vocab()
        self.tmp_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "tmp")
        os.makedirs(self.tmp_dir, exist_ok=True)

    def tearDown(self) -> None:
        shutil.rmtree(self.tmp_dir)

    def test_add_words(self):
        self.undertest.add_words(os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "examples", "vocab_data.txt"))
        self.assertEqual(["house", "dog"], list(self.undertest.vocab.keys()))

    def test_add_word(self):
        self.undertest.add_word("test", cnt=31, vec=[1.42, 1.44, 1.55])
        self.assertEqual(["test"], list(self.undertest.vocab.keys()))
        self.assertTrue("test" in self.undertest)

    def test_count(self):
        self.undertest.add_words(os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "examples", "vocab_data.txt"))
        self.assertEqual(34444, self.undertest.count("house"))

    def test_save_and_load(self):
        self.undertest.add_words(os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "examples", "vocab_data.txt"))
        self.undertest.add_word("test", cnt=31, vec=[1.42, 1.44, 1.55])
        vocab_path = f"{self.tmp_dir}/vocab.dat"
        self.undertest.save(vocab_path)
        vocab = Vocab.load(vocab_path)
        self.assertEqual(["house", "dog", "test"], list(vocab.vocab.keys()))


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, division, print_function

import pickle
import logging

import datasets


_CITATION = """\
@ARTICLE{Kraljevic2021-ln,
  title="Multi-domain clinical natural language processing with {MedCAT}: The Medical Concept Annotation Toolkit",
  author="Kraljevic, Zeljko and Searle, Thomas and Shek, Anthony and Roguski, Lukasz and Noor, Kawsar and Bean, Daniel and Mascio, Aurelie and Zhu, Leilei and Folarin, Amos A and Roberts, Angus and Bendayan, Rebecca and Richardson, Mark P and Stewart, Robert and Shah, Anoop D and Wong, Wai Keong and Ibrahim, Zina and Teo, James T and Dobson, Richard J B",
  journal="Artif. Intell. Med.",
  volume=117,
  pages="102083",
  month=jul,
  year=2021,
  issn="0933-3657",
  doi="10.1016/j.artmed.2021.102083"
}
"""

_DESCRIPTION = """\
Takes as input a pickled dict of annotated documents from MedCAT. The format should be:
    {'document_id': {'entities': <entities>, ...}
Where entities is the output from medcat.get_entities(<...>)['entities']
"""


class MedCATAnnotationsConfig(datasets.BuilderConfig):
    """ BuilderConfig for MedCATAnnotations.

        Args:
          **kwargs: keyword arguments forwarded to super.
    """
    pass


class MedCATAnnotations(datasets.GeneratorBasedBuilder):
    """MedCATAnnotations: Output of MedCAT"""

    BUILDER_CONFIGS = [
        MedCATAnnotationsConfig(
            name="pickle",
            version=datasets.Version("1.0.0", ""),
            description="Pickled output from MedCAT",
        ),
    ]

    def _info(self):
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "id": datasets.Value("int32"),
                    "document_id": datasets.Value("string"),
                    "context_left": datasets.Value("string"),
                    "context_right": datasets.Value("string"),
                    "context_center": datasets.Value("string"),
                }
            ),
            # No default supervised_keys (as we have to pass both question
            # and context as input).
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager): # noqa
        """Returns SplitGenerators."""
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={
                    "filepath": self.config.data_files,
                },
            ),
        ]

    def _generate_examples(self, filepath):
        """This function returns the examples in the raw (text) form."""
        logging.info("generating examples from = %s", filepath)
        with open(filepath, 'rb') as f:
            docs = pickle.load(f)
            for doc_id in docs:
                doc = docs[doc_id]
                for entity_id, entity in doc['entities'].items():
                    yield "{}|{}".format(doc_id, entity['id']), {
                            'id': int(entity_id),
                            'document_id': str(doc_id),
                            'context_left': "".join(entity['context_left']),
                            'context_right': "".join(entity['context_right']),
                            'context_center': "".join(entity['context_center']),
                            }



import getpass
import elasticsearch
import elasticsearch.helpers
from IPython.display import display, HTML
from datetime import datetime

from typing import Dict, Optional


class CogStackConn(object):
    def __init__(self, host, port=9200, username: Optional[str] = None, password: Optional[str] = None, scheme: str='https',
                 timeout: int = 360, max_retries: int = 10, retry_on_timeout: bool = True, **kwargs):
        username, password = self._check_auth_details(username, password)

        self.elastic = elasticsearch.Elasticsearch(hosts=[{'host': host, 'port': port}],
                                         http_auth=(username, password),
                                         scheme=scheme,
                                         verify_certs=False,
                                         timeout=timeout,
                                         max_retries=max_retries,
                                         retry_on_timeout=retry_on_timeout,
                                         **kwargs)

    def _check_auth_details(self, username: Optional[str] = None, password: Optional[str] = None):
        if username is None:
            username = input("Username:")
        if password is None:
            password = getpass.getpass("Password:")

        # TODO: Implement auth check, for now I assume all is fine
        return username, password

    def get_docs_generator(self, query: Dict, index: str, es_gen_size: int=800, request_timeout: int=840000, **kwargs):
        docs_generator = elasticsearch.helpers.scan(self.elastic,
            query=query,
            index=index,
            size=es_gen_size,
            request_timeout=request_timeout,
            **kwargs)

        return docs_generator

    def get_text_for_doc(self, doc_id, index='epr_documents', text_field='body_analysed'):
        r = self.elastic.get(index=index, id=doc_id)
        text = r['_source'][text_field]
        return text

    def show_all_ent_cntx(self, stream, cui: str, cntx_size: int = 100, index='epr_documents', text_field='body_analysed'):
        for id in range(len(stream['entities'])):
            if stream['entities'][id]['conceptId'] == cui:
                print(stream['entities'][id]['name'])
                print("Status: " + stream['entities'][id]['metaSubject'])
                print("Presence: " + stream['entities'][id]['metaPresence'])
                print("Time: " + stream['entities'][id]['metaTime'])
                print("Date: " + str(datetime.fromtimestamp((stream['entities'][id]['timestamp']))))

                self.show_ent_cntx(stream, id, cntx_size, index, text_field)

    def show_ent_cntx(self, stream, id: int, cntx_size=100, index='epr_documents', text_field='body_analysed'):
        doc_id = stream['entities'][id]['documentId']
        text = self.get_text_for_doc(doc_id, index=index, text_field=text_field)

        start = stream['entities'][id]['start']
        c_start = max(0, start-cntx_size)
        end = stream['entities'][id]['end']
        c_end = min(len(text), end+cntx_size)

        ent_cntx = text[c_start:start] + "<span style='background-color: #53f725'>" + text[start:end] + "</span>" + text[end:c_end]
        ent_cntx.replace("\n", "<br />")
        display(HTML(ent_cntx))

        if len(text) < start:
            print("Text of the clincal note corrupted: " + text[0:100])


    def bulk_to_cogstack(self):
        # TODO: look the code made for Nazli/Dan
        pass



from __future__ import absolute_import, division, print_function

import json
import logging

import datasets


_CITATION = """\
@misc{kraljevic2020multidomain,
      title={Multi-domain Clinical Natural Language Processing with MedCAT: the Medical Concept Annotation Toolkit},
      author={Zeljko Kraljevic and Thomas Searle and Anthony Shek and Lukasz Roguski and Kawsar Noor and Daniel Bean and Aurelie Mascio and Leilei Zhu and Amos A Folarin and Angus Roberts and Rebecca Bendayan and Mark P Richardson and Robert Stewart and Anoop D Shah and Wai Keong Wong and Zina Ibrahim and James T Teo and Richard JB Dobson},
      year={2020},
      eprint={2010.01165},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}
"""

_DESCRIPTION = """\
Takes as input a json export from medcattrainer."""


class MedCATAnnotationsConfig(datasets.BuilderConfig):
    """ BuilderConfig for MedCATNER.

        Args:
          **kwargs: keyword arguments forwarded to super.
    """
    pass


class MedCATNER(datasets.GeneratorBasedBuilder):
    """MedCATNER: Output of MedCATtrainer"""

    BUILDER_CONFIGS = [
        MedCATAnnotationsConfig(
            name="json",
            version=datasets.Version("1.0.0", ""),
            description="JSON output from MedCATtrainer",
        ),
    ]


    def _info(self):
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "id": datasets.Value("int32"),
                    "text": datasets.Value("string"),
                    "name": datasets.Value("string"),
                    "ent_starts": datasets.Sequence(datasets.Value("int32")),
                    "ent_ends": datasets.Sequence(datasets.Value("int32")),
                    "ent_cuis": datasets.Sequence(datasets.Value("string")),

                }
            ),
            # No default supervised_keys (as we have to pass both question
            # and context as input).
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager): # noqa
        """Returns SplitGenerators."""
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={
                    "filepaths": self.config.data_files['train'],
                },
            ),
        ]

    def _generate_examples(self, filepaths):
        for filepath in filepaths:
            logging.info("generating examples from = %s", filepath)
            with open(filepath, 'r') as f:
                projects = json.load(f)['projects']
                for project in projects:
                    for ind, doc in enumerate(project['documents']):
                        starts = []
                        ends = []
                        cuis = []
                        for entity in doc['annotations']:
                            if entity.get('correct', True) or \
                               entity.get('manually_created', False) or \
                               entity.get('alternative', False):

                                starts.append(entity['start'])
                                ends.append(entity['end'])
                                cuis.append(entity['cui'])
                        doc_id = doc.get('id', ind)
                        doc_name = doc.get('name')
                        yield "{}".format(doc_id), {
                                'id': int(doc_id),
                                'text': str(doc['text']),
                                'name': str(doc_name),
                                'ent_starts': starts,
                                'ent_ends': ends,
                                'ent_cuis': cuis,
                                }



from typing import Dict, Optional, Tuple, Iterable, List
from medcat.tokenizers.meta_cat_tokenizers import TokenizerWrapperBase


def prepare_from_json(data: Dict,
                      cntx_left: int,
                      cntx_right: int,
                      tokenizer: TokenizerWrapperBase,
                      cui_filter: Optional[set] = None,
                      replace_center: Optional[str] = None,
                      prerequisites: Dict = {},
                      lowercase: bool = True) -> Dict:
    """ Convert the data from a json format into a CSV-like format for training. This function is not very efficient (the one
    working with spacy documents as part of the meta_cat.pipe method is much better). If your dataset is > 1M documents think
    about rewriting this function - but would be strange to have more than 1M manually annotated documents.

    Args:
        data (`dict`):
            Loaded output of MedCATtrainer. If we have a `my_export.json` from MedCATtrainer, than data = json.load(<my_export>).
        cntx_left (`int`):
            Size of context to get from the left of the concept
        cntx_right (`int`):
            Size of context to get from the right of the concept
        tokenizer (`medcat.tokenizers.meta_cat_tokenizers`):
            Something to split text into tokens for the LSTM/BERT/whatever meta models.
        replace_center (`str`, optional):
            If not None the center word (concept) will be replaced with whatever this is.
        prerequisites (`dict`, optional):
            A map of prerequisities, for example our data has two meta-annotations (experiencer, negation). Assume I want to create
            a dataset for `negation` but only in those cases where `experiencer=patient`, my prerequisites would be:
                {'Experiencer': 'Patient'} - Take care that the CASE has to match whatever is in the data
        lowercase (`bool`, defaults to True):
            Should the text be lowercased before tokenization

    Returns:
        out_data (`dict`):
            Example: {'category_name': [('<category_value>', '<[tokens]>', '<center_token>'), ...], ...}
    """
    out_data: Dict = {}

    for project in data['projects']:
        for document in project['documents']:
            text = str(document['text'])
            if lowercase:
                text = text.lower()

            if len(text) > 0:
                doc_text = tokenizer(text)

                for ann in document.get('annotations', document.get('entities', {}).values()): # A hack to suport entities and annotations
                    cui = ann['cui']
                    skip = False
                    if 'meta_anns' in ann and prerequisites:
                        # It is possible to require certain meta_anns to exist and have a specific value
                        for meta_ann in prerequisites:
                            if meta_ann not in ann['meta_anns'] or ann['meta_anns'][meta_ann]['value'] != prerequisites[meta_ann]:
                                # Skip this annotation as the prerequisite is not met
                                skip = True
                                break

                    if not skip and (cui_filter is None or not cui_filter or cui in cui_filter):
                        if ann.get('validated', True) and (not ann.get('deleted', False) and not ann.get('killed', False)
                                                           and not ann.get('irrelevant', False)):
                            start = ann['start']
                            end = ann['end']

                            # Get the index of the center token
                            ind = 0
                            for ind, pair in enumerate(doc_text['offset_mapping']):
                                if start >= pair[0] and start < pair[1]:
                                    break

                            _start = max(0, ind - cntx_left)
                            _end = min(len(doc_text['input_ids']), ind + 1 + cntx_right)
                            tkns = doc_text['input_ids'][_start:_end]
                            cpos = cntx_left + min(0, ind-cntx_left)

                            if replace_center is not None:
                                if lowercase:
                                    replace_center = replace_center.lower()
                                for p_ind, pair in enumerate(doc_text['offset_mapping']):
                                    if start >= pair[0] and start < pair[1]:
                                        s_ind = p_ind
                                    if end > pair[0] and end <= pair[1]:
                                        e_ind = p_ind

                                ln = e_ind - s_ind
                                tkns = tkns[:cpos] + tokenizer(replace_center)['input_ids'] + tkns[cpos+ln+1:]

                            # Backward compatibility if meta_anns is a list vs dict in the new approach
                            meta_anns = []
                            if 'meta_anns' in ann:
                                meta_anns = ann['meta_anns'].values() if type(ann['meta_anns']) == dict else ann['meta_anns']

                            # If the annotation is validated
                            for meta_ann in meta_anns:
                                name = meta_ann['name']
                                value = meta_ann['value']

                                sample = [tkns, cpos, value]

                                if name in out_data:
                                    out_data[name].append(sample)
                                else:
                                    out_data[name] = [sample]
    return out_data


def encode_category_values(data: Dict, existing_category_value2id: Optional[Dict] = None) -> Tuple:
    r''' Converts the category values in the data outputed by `prepare_from_json`
    into integere values.

    Args:
        data (`dict`):
            Output of `prepare_from_json`
        existing_category_value2id(`dict`, optional):
            Map from category_value to id (old/existing)

    Returns:
        data (`dict`):
            New data with integeres inplace of strings for categry values.
        category_value2id (`dict`):
            Map rom category value to ID for all categories in the data.
    '''
    data = list(data)
    if existing_category_value2id is not None:
        category_value2id = existing_category_value2id
    else:
        category_value2id = {}

    category_values = set([x[2] for x in data])
    for c in category_values:
        if c not in category_value2id:
            category_value2id[c] = len(category_value2id)

    # Map values to numbers
    for i in range(len(data)):
        data[i][2] = category_value2id[data[i][2]]

    return data, category_value2id


def json_to_fake_spacy(data: Dict, id2text: Dict) -> Iterable:
    r''' Creates a generator of fake spacy documents, used for running
    meta_cat pipe separately from main cat pipeline.

    Args:
        data(`dict`):
            Output from cat formated as: {<id>: <output of get_entities, ...}
        id2text(`dict`):
            Map from document id to text of that documetn

    Returns:
        generator:
            Generator of spacy like documents that can be feed into meta_cat.pipe
    '''
    for id_ in data.keys():
        ents = data[id_]['entities'].values()

        doc = Doc(text=id2text[id_], id_=id_)
        doc.ents.extend([Span(ent['start'], ent['end'], ent['id']) for ent in ents])

        yield doc


class Empty(object):
    def __init__(self) -> None:
        pass


class Span(object):
    def __init__(self, start_char: str, end_char: str, id_: str) -> None:
        self._ = Empty()
        self.start_char = start_char
        self.end_char = end_char
        self._.id = id_  # type: ignore
        self._.meta_anns = None # type: ignore


class Doc(object):
    def __init__(self, text: str, id_: str) -> None:
        self._ = Empty()
        self._.share_tokens = None  # type: ignore
        self.ents: List = []
        # We do not have overlapps at this stage
        self._ents = self.ents
        self.text = text
        self.id = id_



import os
import shutil
import unittest

from transformers import AutoTokenizer

from medcat.meta_cat import MetaCAT
from medcat.config_meta_cat import ConfigMetaCAT
from medcat.tokenizers.meta_cat_tokenizers import TokenizerWrapperBERT


class CATTests(unittest.TestCase):

    @classmethod
    def setUpClass(self) -> None:
        tokenizer = TokenizerWrapperBERT(AutoTokenizer.from_pretrained("bert-base-uncased"))
        config = ConfigMetaCAT()
        config.general['category_name'] = 'Status'
        config.general['cntx_left'] = 15
        config.general['cntx_right'] = 10
        config.general['seed'] = 13
        config.train['nepochs'] = 10
        config.model['input_size'] = 300
        config.train['auto_save_model'] = True

        self.meta_cat = MetaCAT(tokenizer=tokenizer, embeddings=None, config=config)

        self.tmp_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "tmp")
        os.makedirs(self.tmp_dir, exist_ok=True)


    def tearDown(self) -> None:
        shutil.rmtree(self.tmp_dir)


    def test_train(self):
        json_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'resources', 'mct_export_for_meta_cat_test.json')
        results = self.meta_cat.train(json_path, save_dir_path=self.tmp_dir)

        self.assertEqual(results['f1'], 1.0)


    def test_save_load(self):
        json_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'resources', 'mct_export_for_meta_cat_test.json')
        self.meta_cat.train(json_path, save_dir_path=self.tmp_dir)
        self.meta_cat.save(self.tmp_dir)
        n_meta_cat = MetaCAT.load(self.tmp_dir)

        f1 = self.meta_cat.eval(json_path)['f1']
        n_f1 = n_meta_cat.eval(json_path)['f1']

        self.assertEqual(f1, n_f1)


if __name__ == '__main__':
    unittest.main()



import os
import pandas as pd


def get_index_queries():
    r''' Run before everything to speed up things
    '''
    return ['CREATE INDEX patientId FOR (p:Patient) ON (p.patientId);',
            'CREATE INDEX conceptId FOR (c:Concept) ON (c.conceptId);',
            'CREATE INDEX documentId FOR (d:Document) ON (d.documentId);']


def create_neo_csv(data, columns, output_dir='/etc/lib/neo4j/import/',
                   base_name='patients'):
    r''' Creates a patients CSV for neo4j load csv function

    Args:
        data:
            A dataframe or path to a dataframe with the required data
        columns:
            What data to use from the dataframe
        output_dir:
            Where to save the CSVs, should be the neo4j imports path if possible
        base_name:
            Name of the csv
    '''
    if isinstance(data, pd.DataFrame):
        df = data
    else:
        df = pd.read_csv(data)

    # Remove duplicates
    df = df.drop_duplicates(subset=columns)

    out_df = df[columns]
    data_path = os.path.join(output_dir, f"{base_name}.csv")
    out_df.to_csv(data_path, index=False)


def create_patients_csv(data, output_dir='/etc/lib/neo4j/import/',
                        base_name='patients'):
    r''' Creates a patients CSV for neo4j load csv function

    Args:
        data:
            A dataframe or path to a dataframe with the required data: patientId,
            sex, ethnicity, dob
        output_dir:
            Where to save the CSVs, should be the neo4j imports path if possible,
            but writing there could be only admin
    '''
    query = (
        'USING PERIODIC COMMIT 100000 \n'
        f'LOAD CSV WITH HEADERS FROM  "file:///{base_name}.csv" AS row \n'
        'CREATE (:Patient {patientId: toString(row.patientId), \n'
        '                  sex: toString(row.sex), \n'
        '                  ethnicity: toString(row.ethnicity), \n'
        '                  dob: datetime(row.dob)}) \n'
        )

    create_neo_csv(data=data, columns=['patientId', 'sex', 'ethnicity', 'dob'],
                   output_dir=output_dir, base_name=base_name)

    return query


def create_documents_csv(data, output_dir='/etc/lib/neo4j/import/',
                         base_name='documents'):
    r''' Creates a patients CSV for neo4j load csv function

    Args:
        data:
            A dataframe or path to a dataframe with the required data: documentId
        output_dir:
            Where to save the CSVs, should be the neo4j imports path if possible
    '''
    query = (
        'USING PERIODIC COMMIT 100000 \n'
        f'LOAD CSV WITH HEADERS FROM  "file:///{base_name}.csv" AS row \n'
        'CREATE (:Document {documentId: toString(row.documentId)}) \n'
        )

    create_neo_csv(data=data, columns=['documentId'],
                   output_dir=output_dir, base_name=base_name)

    return query


def create_concepts_csv(data, output_dir='/etc/lib/neo4j/import/',
                         base_name='concepts'):
    r''' Creates a patients CSV for neo4j load csv function

    Args:
        data:
            A dataframe or path to a dataframe with the required data: conceptId,
            name and type
        output_dir:
            Where to save the CSVs, should be the neo4j imports path if possible
    '''
    query = (
        'USING PERIODIC COMMIT 100000 \n'
        f'LOAD CSV WITH HEADERS FROM  "file:///{base_name}.csv" AS row \n'
        'CREATE (:Concept {conceptId: toString(row.conceptId), \n'
        '                  type: toString(row.type), \n'
        '                  name: toString(row.name)}) \n'
        )

    create_neo_csv(data=data, columns=['conceptId', 'name', 'type'],
                   output_dir=output_dir, base_name=base_name)

    return query


def create_document2patient_csv(data, output_dir='/etc/lib/neo4j/import/',
                                base_name='document2patient'):

    r''' Creates a patients CSV for neo4j load csv function

    Args:
        data:
            A dataframe or path to a dataframe with the required data: patientId and
            documentId
        output_dir:
            Where to save the CSVs, should be the neo4j imports path if possible
    '''
    query = (
        'USING PERIODIC COMMIT 100000 \n'
        f'LOAD CSV WITH HEADERS FROM  "file:///{base_name}.csv" AS row \n'
        'MATCH (pt:Patient {patientId: toString(row.patientId)}) \n'
        'MATCH (doc:Document {documentId: toString(row.documentId)}) \n'
        'CREATE (pt)-[:HAS]->(doc); \n'
        )

    create_neo_csv(data=data, columns=['patientId', 'documentId'],
                   output_dir=output_dir, base_name=base_name)

    return query


def create_concept_ontology_csv(data, output_dir='/etc/lib/neo4j/import/',
                                base_name='concept_ontology'):

    r''' Creates a patients CSV for neo4j load csv function

    Args:
        data:
            A dataframe or path to a dataframe with the required data: child, parent
        output_dir:
            Where to save the CSVs, should be the neo4j imports path if possible
    '''
    query = (
        'USING PERIODIC COMMIT 100000 \n'
        f'LOAD CSV WITH HEADERS FROM  "file:///{base_name}.csv" AS row \n'
        'MATCH (child:Concept {conceptId: toString(row.child)}) \n'
        'MATCH (parent:Concept {conceptId: toString(row.parent)}) \n'
        'CREATE (child)-[:IS_A]->(parent); \n'
        )

    create_neo_csv(data=data, columns=['child', 'parent'],
                   output_dir=output_dir, base_name=base_name)

    return query


def create_document2concept_csv(data, output_dir='/etc/lib/neo4j/import/',
                         base_name='document2concepts'):
    r''' Creates a patients CSV for neo4j load csv function

    Args:
        data:
            A dataframe or path to a dataframe with the required data: 'conceptId',
            'documentId', 'contextSimilarity', 'start', 'end', 'timestamp',
            'metaSubject', 'metaPresence', 'metaTime'
        output_dir:
            Where to save the CSVs, should be the neo4j imports path if possible
    '''
    query = (
        'USING PERIODIC COMMIT 100000 \n'
        f'LOAD CSV WITH HEADERS FROM  "file:///{base_name}.csv" AS row \n'
        'MATCH (doc:Document{documentId: toString(row.documentId)}) \n'
        'MATCH (concept:Concept {conceptId: toString(row.conceptId)}) \n'
        'CREATE (doc)-[:HAS {start: toInteger(row.start), \n'
        '                   end: toInteger(row.end), \n'
        '                   timestamp: toInteger(row.timestamp), \n'
        '                   contextSimilarity: toFloat(row.contextSimilarity), \n'
        '                   metaSubject: toString(row.metaSubject), \n'
        '                   metaPresence: toString(row.metaPresence), \n'
        '                   metaTime: toString(row.metaTime) \n'
        '            }]->(concept); \n'
        )

    columns = ['conceptId', 'documentId', 'contextSimilarity', 'start',
                'end', 'timestamp', 'metaSubject', 'metaPresence', 'metaTime']

    create_neo_csv(data=data, columns=columns,
                   output_dir=output_dir, base_name=base_name)

    return query


def get_data_from_docs(docs, doc2pt, doc2time=None):
    data = [['conceptId', 'documentId', 'contextSimilarity',
             'start', 'end', 'timestamp', 'metaSubject',
             'metaPresence', 'metaTime']]

    for doc_id, doc in docs.items():
        row = []
        for ent in doc['entities'].values():
            #if ent['meta_anns']['Subject']['value'] == 'Patient' and \
            #   ent['meta_anns']['Presence']['value'] == 'True':
            if doc2time is not None:
                t = doc2time[doc_id]
            else:
                t = ent['document_timestamp']

            row = [ent['cui'], doc_id,
                   ent['context_similarity'],
                   ent['start'], ent['end'],
                   t,
                   ent['meta_anns'].get('Subject', {}).get('value', None),
                   ent['meta_anns'].get('Presence', {}).get('value', None),
                   ent['meta_anns'].get('Time', {}).get('value', None)]
            data.append(row)
            row = []

    return data



def get_lr_linking(config, cui_count):
    if config.linking['optim']['type'] == 'standard':
        return config.linking['optim']['lr']
    elif config.linking['optim']['type'] == 'linear':
        lr = config.linking['optim']['base_lr']
        cui_count += 1 # Just in case incrase by 1
        return max(lr / cui_count, config.linking['optim']['min_lr'])
    else:
        raise Exception("Optimizer not implemented")


def get_batch(ind, batch_size, x, y, cpos, device):
    # Get the start/end index for this batch
    start = ind * batch_size
    end = (ind+1) * batch_size

    # Get the batch
    x_batch = x[start:end]
    y_batch = y[start:end]
    c_batch = cpos[start:end]

    # Return and move the batches to the right device
    return x_batch.to(device), y_batch.to(device), c_batch.to(device)


def load_hf_tokenizer(tokenizer_name):
    try:
        from transformers import AutoTokenizer
        hf_tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
    except Exception:
        # Where is log defined?
        log.exception("The Huggingface tokenizer could not be created") # noqa

    return hf_tokenizer


def build_vocab_from_hf(model_name, hf_tokenizer, vocab):
    rebuild = False
    # Check is it necessary
    for i in range(hf_tokenizer.vocab_size):
        tkn = hf_tokenizer.ids_to_tokens[i]
        if tkn not in vocab:
            rebuild = True

    if rebuild:
        # Where is log defined?
        log.info("Rebuilding vocab") # noqa
        try:
            from transformers import AutoModel
            model = AutoModel.from_pretrained(model_name)
            if 'xlnet' in model_name.lower():
                embs = model.get_input_embeddings().weight.cpu().detach().numpy()
            else:
                embs = model.embeddings.word_embeddings.weight.cpu().detach().numpy()

            # Reset all vecs in current vocab
            vocab.vec_index2word = {}
            for ind in vocab.index2word.keys():
                vocab.vocab[vocab.index2word[ind]]['vec'] = None

            for i in range(hf_tokenizer.vocab_size):
                tkn = hf_tokenizer.ids_to_tokens[i]
                vec = embs[i]
                vocab.add_word(word=tkn, vec=vec, replace=True)

            # Crate the new unigram table
            vocab.reset_counts()
            vocab.make_unigram_table()
        except Exception:
            # Where is log defined?
            log.exception("The Huggingface model could not be loaded") # noqa



from medcat.utils.medmentions import original2concept_csv
from medcat.utils.medmentions import original2json
from medcat.utils.medmentions import original2pure_text

_ = original2json("../../examples/medmentions/medmentions.txt", '../../examples/medmentions/tmp_medmentions.json')
_ = original2concept_csv("../../examples/medmentions/medmentions.txt", '../../examples/medmentions/tmp_medmentions.csv')
original2pure_text("../../examples/medmentions/medmentions.txt", '../../examples/medmentions/tmp_medmentions_text_only.txt')



import os
from abc import ABC, abstractmethod
from typing import List, Dict, Optional, Union, overload
from tokenizers import Tokenizer, ByteLevelBPETokenizer
from transformers.models.bert.tokenization_bert_fast import BertTokenizerFast


class TokenizerWrapperBase(ABC):

    name: str

    def __init__(self, hf_tokenizer: Optional[Tokenizer] = None) -> None:
        self.hf_tokenizers = hf_tokenizer

    @overload
    def __call__(self, text: str) -> Dict: ...

    @overload
    def __call__(self, text: List[str]) -> List[Dict]: ...

    @abstractmethod
    def __call__(self, text: Union[str, List[str]]) -> Union[Dict, List[Dict]]: ...

    @abstractmethod
    def save(self, dir_path: str) -> None: ...

    @classmethod
    @abstractmethod
    def load(cls, dir_path: str, **kwargs) -> Tokenizer: ...

    @abstractmethod
    def get_size(self) -> int: ...

    @abstractmethod
    def token_to_id(self, token: str) -> Union[int, List[int]]: ...

    @abstractmethod
    def get_pad_id(self) -> Union[Optional[int], List[int]]: ...

    def ensure_tokenizer(self) -> Tokenizer:
        if self.hf_tokenizers is None:
            raise ValueError("The tokenizer is not loaded yet")
        return self.hf_tokenizers


class TokenizerWrapperBPE(TokenizerWrapperBase):
    ''' Wrapper around a huggingface tokenizer so that it works with the
    MetaCAT models.

    Args:
        hf_tokenizers (`tokenizers.ByteLevelBPETokenizer`):
            A huggingface BBPE tokenizer.
    '''
    name = 'bbpe'

    def __init__(self, hf_tokenizers: Optional[ByteLevelBPETokenizer] = None) -> None:
        super().__init__(hf_tokenizers)

        if self.hf_tokenizers is not None:
            # For whatever reason added tokens do not persist with this tokenizer, what to do
            self.hf_tokenizers.add_tokens(['<PAD>'])

    @overload
    def __call__(self, text: str) -> Dict: ...

    @overload
    def __call__(self, text: List[str]) -> List[Dict]: ...

    def __call__(self, text: Union[str, List[str]]) -> Union[Dict, List[Dict]]:
        r''' Tokenize some text

        Args:
            text (`Union(str, List[str])`):
                Text/texts to be tokenized.

        Returns:
            res (`Union(dict, List[dict])`):
                Dictionary/ies containing `offset_mapping`, `input_ids` and `tokens` corresponding to the
                input text/s.

        '''
        self.hf_tokenizers = self.ensure_tokenizer()

        if isinstance(text, str):
            result = self.hf_tokenizers.encode(text)

            return {'offset_mapping': result.offsets,
                    'input_ids': result.ids,
                    'tokens': result.tokens,
                    }
        elif isinstance(text, list):
            results = self.hf_tokenizers.encode_batch(text)
            output = []
            for result in results:
                output.append({'offset_mapping': result.offsets,
                    'input_ids': result.ids,
                    'tokens': result.tokens,
                })

            return output
        else:
            raise Exception("Unsuported input type, supported: text/list, but got: {}".format(type(text)))

    def save(self, dir_path: str) -> None:
        self.hf_tokenizers = self.ensure_tokenizer()

        if self.hf_tokenizers is None:
            raise ValueError("The tokenizer is not loaded yet")

        self.hf_tokenizers.save_model(dir_path, prefix=self.name)

    @classmethod
    def load(cls, dir_path: str, **kwargs) -> "TokenizerWrapperBPE":
        tokenizer = cls()
        vocab_file = os.path.join(dir_path, f'{tokenizer.name}-vocab.json')
        merges_file = os.path.join(dir_path, f'{tokenizer.name}-merges.txt')
        tokenizer.hf_tokenizers = ByteLevelBPETokenizer.from_file(vocab_filename=vocab_file,
                                                                  merges_filename=merges_file,
                                                                  **kwargs)
        # For whatever reason added tokens do not persist with this tokenizer, so we added it at each load
        tokenizer.hf_tokenizers.add_tokens(['<PAD>'])
        return tokenizer

    def get_size(self) -> int:
        self.hf_tokenizers = self.ensure_tokenizer()
        return self.hf_tokenizers.get_vocab_size()

    def token_to_id(self, token: str) -> Union[int, List[int]]:
        self.hf_tokenizers = self.ensure_tokenizer()
        return self.hf_tokenizers.token_to_id(token)

    def get_pad_id(self) -> Union[int, List[int]]:
        pad = self.token_to_id('<PAD>')
        if pad is None:
            raise Exception("No <PAD> token in the vocabulary of the tokenizer, please add it")
        return pad


class TokenizerWrapperBERT(TokenizerWrapperBase):
    ''' Wrapper around a huggingface BERT tokenizer so that it works with the
    MetaCAT models.

    Args:
        hf_tokenizers (`transformers.models.bert.tokenization_bert_fast.BertTokenizerFast`):
            A huggingface Fast BERT.
    '''
    name = 'bert-tokenizer'

    def __init__(self, hf_tokenizers: Optional[BertTokenizerFast] = None) -> None:
        super().__init__(hf_tokenizers)

    @overload
    def __call__(self, text: str) -> Dict: ...

    @overload
    def __call__(self, text: List[str]) -> List[Dict]: ...

    def __call__(self, text: Union[str, List[str]]) -> Union[Dict, List[Dict]]:
        self.hf_tokenizers = self.ensure_tokenizer()
        if isinstance(text, str):
            result = self.hf_tokenizers.encode_plus(text, return_offsets_mapping=True,
                    add_special_tokens=False)

            return {'offset_mapping': result['offset_mapping'],
                    'input_ids': result['input_ids'],
                    'tokens':  self.hf_tokenizers.convert_ids_to_tokens(result['input_ids']),
                    }
        elif isinstance(text, list):
            results = self.hf_tokenizers._batch_encode_plus(text, return_offsets_mapping=True,
                    add_special_tokens=False)
            output = []
            for ind in range(len(results['input_ids'])):
                output.append({'offset_mapping': results['offset_mapping'][ind],
                    'input_ids': results['input_ids'][ind],
                    'tokens':  self.hf_tokenizers.convert_ids_to_tokens(results['input_ids'][ind]),
                    })
            return output
        else:
            raise Exception("Unsuported input type, supported: text/list, but got: {}".format(type(text)))

    def save(self, dir_path: str) -> None:
        self.hf_tokenizers = self.ensure_tokenizer()
        path = os.path.join(dir_path, self.name)
        self.hf_tokenizers.save_pretrained(path)

    @classmethod
    def load(cls, dir_path: str, **kwargs) -> "TokenizerWrapperBERT":
        tokenizer = cls()
        path = os.path.join(dir_path, cls.name)
        tokenizer.hf_tokenizers = BertTokenizerFast.from_pretrained(path, **kwargs)

        return tokenizer

    def get_size(self) -> int:
        self.hf_tokenizers = self.ensure_tokenizer()
        return len(self.hf_tokenizers.vocab)

    def token_to_id(self, token: str) -> Union[int, List[int]]:
        self.hf_tokenizers = self.ensure_tokenizer()
        return self.hf_tokenizers.convert_tokens_to_ids(token)

    def get_pad_id(self) -> Optional[int]:
        self.hf_tokenizers = self.ensure_tokenizer()
        return self.hf_tokenizers.pad_token_id



from spacy.language import Language
from spacy.tokens import Doc
from medcat.config import Config
from medcat.pipeline.pipe_runner import PipeRunner


def tag_skip_and_punct(nlp: Language, name: str, config: Config) -> "_Tagger":
    r''' Detects and tags spacy tokens that are punctuation and that should be skipped.

     Args:
         nlp (spacy.language.<lng>):
             The base spacy NLP pipeline.
         name (`str`):
             The component instance name.
         config (`medcat.config.Config`):
             Global config for medcat.
    '''
    return _Tagger(nlp, name, config)


# This is not elegant.
tag_skip_and_punct.name = "tag_skip_and_punct"  # type: ignore


class _Tagger(PipeRunner):

    # Override
    def __init__(self, nlp: Language, name: str, config: Config) -> None:
        self.name = name
        self.config = config
        super().__init__(self.config.general['workers'])

    # Override
    def __call__(self, doc: Doc) -> Doc:
        # Make life easier
        cnf_p = self.config.preprocessing

        for token in doc:
            if self.config.punct_checker.match(token.lower_) and token.text not in cnf_p['keep_punct']:
                # There can't be punct in a token if it also has text
                token._.is_punct = True
                token._.to_skip = True
            elif self.config.word_skipper.match(token.lower_):
                # Skip if specific strings
                token._.to_skip = True
            elif cnf_p['skip_stopwords'] and token.is_stop:
                token._.to_skip = True

        return doc



import os
import json
import re
import hashlib
import pandas as pd


def parse_file(filename, first_row_header=True, columns=None):
    with open(filename, encoding='utf-8') as f:
        entities = [[n.strip() for n in line.split('\t')] for line in f]
        return pd.DataFrame(entities[1:], columns=entities[0] if first_row_header else columns)


class Snomed:
    r"""
    Pre-process SNOMED CT release files:
    Args:
        data_path:
            Path to the unzipped SNOMED CT folder
        uk_ext: bool
            Specification of a SNOMED UK extension after 2021 to process the divergent release format.
    """

    def __init__(self, data_path, uk_ext=False):
        self.data_path = data_path
        self.release = data_path[-16:-8]
        self.uk_ext = uk_ext

    def to_concept_df(self, ):
        """
        Please remember to specify if the version is a SNOMED UK extension released after 2021.
        This can be done prior to this step: Snomed.uk_ext = True.
        This step is not required for UK extension releases pre-2021.

        :return: SNOMED CT concept DataFrame ready for MEDCAT CDB creation
        """
        snomed_releases = []
        paths = []
        if "Snapshot" in os.listdir(self.data_path):
            paths.append(self.data_path)
            snomed_releases.append(self.release)
        else:
            for folder in os.listdir(self.data_path):
                if "SnomedCT" in folder:
                    paths.append(os.path.join(self.data_path, folder))
                    snomed_releases.append(folder[-16:-8])
        if len(paths) == 0:
            raise FileNotFoundError('Incorrect path to SNOMED CT directory')

        df2merge = []
        for i, snomed_release in enumerate(snomed_releases):
            contents_path = os.path.join(paths[i], "Snapshot", "Terminology")
            concept_snapshot = "sct2_Concept_Snapshot"
            description_snapshot = "sct2_Description_Snapshot-en"
            if self.uk_ext:
                if "SnomedCT_UKClinicalRF2_PRODUCTION" in paths[i]:
                    concept_snapshot = "sct2_Concept_UKCLSnapshot"
                    description_snapshot = "sct2_Description_UKCLSnapshot-en"
                elif "SnomedCT_UKEditionRF2_PRODUCTION" in paths[i]:
                    concept_snapshot = "sct2_Concept_UKEDSnapshot"
                    description_snapshot = "sct2_Description_UKEDSnapshot-en"
                elif "SnomedCT_UKClinicalRefsetsRF2_PRODUCTION" in paths[i]:
                    continue
                else:
                    pass

            for f in os.listdir(contents_path):
                m = re.search(f'{concept_snapshot}'+r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)

            int_terms = parse_file(f'{contents_path}/{concept_snapshot}_{snomed_v}_{snomed_release}.txt')
            active_terms = int_terms[int_terms.active == '1']
            del int_terms

            int_desc = parse_file(f'{contents_path}/{description_snapshot}_{snomed_v}_{snomed_release}.txt')
            active_descs = int_desc[int_desc.active == '1']
            del int_desc

            _ = pd.merge(active_terms, active_descs, left_on=['id'], right_on=['conceptId'], how='inner')
            del active_terms
            del active_descs

            active_with_primary_desc = _[_['typeId'] == '900000000000003001']  # active description
            active_with_synonym_desc = _[_['typeId'] == '900000000000013009']  # active synonym
            del _
            active_with_all_desc = pd.concat([active_with_primary_desc, active_with_synonym_desc])

            active_snomed_df = active_with_all_desc[['id_x', 'term', 'typeId']]
            del active_with_all_desc

            active_snomed_df = active_snomed_df.rename(columns={'id_x': 'cui', 'term': 'name', 'typeId': 'name_status'})
            active_snomed_df['ontologies'] = 'SNOMED-CT'
            active_snomed_df['name_status'] = active_snomed_df['name_status'].replace(
                ['900000000000003001', '900000000000013009'],
                ['P', 'A'])
            active_snomed_df = active_snomed_df.reset_index(drop=True)

            temp_df = active_snomed_df[active_snomed_df['name_status'] == 'P'][['cui', 'name']]
            temp_df['description_type_ids'] = temp_df['name'].str.extract(r"\((\w+\s?.?\s?\w+.?\w+.?\w+.?)\)$")
            active_snomed_df = pd.merge(active_snomed_df, temp_df.loc[:, ['cui', 'description_type_ids']],
                                        on='cui',
                                        how='left')
            del temp_df

            # Hash semantic tag to get a 8 digit type_id code
            active_snomed_df['type_ids'] = active_snomed_df['description_type_ids'].apply(
                lambda x: int(hashlib.sha256(str(x).encode('utf-8')).hexdigest(), 16) % 10 ** 8)
            df2merge.append(active_snomed_df)

        return pd.concat(df2merge).reset_index(drop=True)

    def list_all_relationships(self):
        """
        SNOMED CT provides a rich set of inter-relationships between concepts.

        :return: List of all SNOMED CT relationships
        """
        snomed_releases = []
        paths = []
        if "Snapshot" in os.listdir(self.data_path):
            paths.append(self.data_path)
            snomed_releases.append(self.release)
        else:
            for folder in os.listdir(self.data_path):
                if "SnomedCT" in folder:
                    paths.append(os.path.join(self.data_path, folder))
                    snomed_releases.append(folder[-16:-8])
        if len(paths) == 0:
            raise FileNotFoundError('Incorrect path to SNOMED CT directory')

        all_rela = []
        for i, snomed_release in enumerate(snomed_releases):
            contents_path = os.path.join(paths[i], "Snapshot", "Terminology")
            concept_snapshot = "sct2_Concept_Snapshot"
            relationship_snapshot = "sct2_Relationship_Snapshot"
            if self.uk_ext:
                if "SnomedCT_InternationalRF2_PRODUCTION" in paths[i]:
                    concept_snapshot = "sct2_Concept_Snapshot"
                    relationship_snapshot = "sct2_Relationship_Snapshot"
                elif "SnomedCT_UKClinicalRF2_PRODUCTION" in paths[i]:
                    concept_snapshot = "sct2_Concept_UKCLSnapshot"
                    relationship_snapshot = "sct2_Relationship_UKCLSnapshot"
                elif "SnomedCT_UKEditionRF2_PRODUCTION" in paths[i]:
                    concept_snapshot = "sct2_Concept_UKEDSnapshot"
                    relationship_snapshot = "sct2_Relationship_UKEDSnapshot"
                elif "SnomedCT_UKClinicalRefsetsRF2_PRODUCTION" in paths[i]:
                    continue
                else:
                    pass

            for f in os.listdir(contents_path):
                m = re.search(f'{concept_snapshot}'+r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)
            int_relat = parse_file(f'{contents_path}/{relationship_snapshot}_{snomed_v}_{snomed_release}.txt')
            active_relat = int_relat[int_relat.active == '1']
            del int_relat

            all_rela.extend([relationship for relationship in active_relat["typeId"].unique()])
        return all_rela

    def relationship2json(self, relationshipcode, output_jsonfile):
        """

        :param relationshipcode: A single SCTID or unique concept identifier of the relationship type
        :param output_jsonfile: Name of json file output. Tip: include SNOMED edition to avoid downstream confusions
        :return: json file  of relationship mapping
        """
        snomed_releases = []
        paths = []
        if "Snapshot" in os.listdir(self.data_path):
            paths.append(self.data_path)
            snomed_releases.append(self.release)
        else:
            for folder in os.listdir(self.data_path):
                if "SnomedCT" in folder:
                    paths.append(os.path.join(self.data_path, folder))
                    snomed_releases.append(folder[-16:-8])
        if len(paths) == 0:
            raise FileNotFoundError('Incorrect path to SNOMED CT directory')

        output_dict = {}
        for i, snomed_release in enumerate(snomed_releases):
            contents_path = os.path.join(paths[i], "Snapshot", "Terminology")
            concept_snapshot = "sct2_Concept_Snapshot"
            relationship_snapshot = "sct2_Relationship_Snapshot"
            if self.uk_ext:
                if "SnomedCT_InternationalRF2_PRODUCTION" in paths[i]:
                    concept_snapshot = "sct2_Concept_Snapshot"
                    relationship_snapshot = "sct2_Relationship_Snapshot"
                elif "SnomedCT_UKClinicalRF2_PRODUCTION" in paths[i]:
                    concept_snapshot = "sct2_Concept_UKCLSnapshot"
                    relationship_snapshot = "sct2_Relationship_UKCLSnapshot"
                elif "SnomedCT_UKEditionRF2_PRODUCTION" in paths[i]:
                    concept_snapshot = "sct2_Concept_UKEDSnapshot"
                    relationship_snapshot = "sct2_Relationship_UKEDSnapshot"
                elif "SnomedCT_UKClinicalRefsetsRF2_PRODUCTION" in paths[i]:
                    continue
                else:
                    pass
            for f in os.listdir(contents_path):
                m = re.search(f'{concept_snapshot}'+r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)
            int_relat = parse_file(f'{contents_path}/{relationship_snapshot}_{snomed_v}_{snomed_release}.txt')
            active_relat = int_relat[int_relat.active == '1']
            del int_relat

            relationship = dict([(key, []) for key in active_relat["destinationId"].unique()])
            for _, v in active_relat.iterrows():
                if v['typeId'] == str(relationshipcode):
                    _ = v['destinationId']
                    relationship[_].append(v['sourceId'])
                else:
                    pass
            output_dict.update(relationship)
        with open(output_jsonfile, 'w') as json_file:
            json.dump(output_dict, json_file)
        return

    def map_snomed2icd10(self):
        """

        :return: SNOMED to ICD10 mapping DataFrame which includes all metadata
        """
        snomed_releases = []
        paths = []
        if "Snapshot" in os.listdir(self.data_path):
            paths.append(self.data_path)
            snomed_releases.append(self.release)
        else:
            for folder in os.listdir(self.data_path):
                if "SnomedCT" in folder:
                    paths.append(os.path.join(self.data_path, folder))
                    snomed_releases.append(folder[-16:-8])
        if len(paths) == 0:
            raise FileNotFoundError('Incorrect path to SNOMED CT directory')
        df2merge = []
        for i, snomed_release in enumerate(snomed_releases):
            refset_terminology = f'{paths[i]}/Snapshot/Refset/Map'
            icd10_ref_set = 'der2_iisssccRefset_ExtendedMapSnapshot'
            if self.uk_ext:
                if "SnomedCT_InternationalRF2_PRODUCTION" in paths[i]:
                    icd10_ref_set = "der2_iisssccRefset_ExtendedMapSnapshot"
                elif "SnomedCT_UKClinicalRF2_PRODUCTION" in paths[i]:
                    icd10_ref_set = "der2_iisssccRefset_ExtendedMapUKCLSnapshot"
                elif "SnomedCT_UKEditionRF2_PRODUCTION" in paths[i]:
                    icd10_ref_set = "der2_iisssccRefset_ExtendedMapUKEDSnapshot"
                elif "SnomedCT_UKClinicalRefsetsRF2_PRODUCTION" in paths[i]:
                    continue
                else:
                    pass
            for f in os.listdir(refset_terminology):
                m = re.search(f'{icd10_ref_set}'+r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)
            mappings = parse_file(f'{refset_terminology}/{icd10_ref_set}_{snomed_v}_{snomed_release}.txt')
            mappings = mappings[mappings.active == '1']
            icd_mappings = mappings.sort_values(by=['referencedComponentId', 'mapPriority', 'mapGroup']).reset_index(
                drop=True)
            df2merge.append(icd_mappings)
        return pd.concat(df2merge)

    def map_snomed2opcs4(self):
        """

        :return: SNOMED to OPSC4 mapping DataFrame which includes all metadata
        """
        snomed_releases = []
        paths = []
        if "Snapshot" in os.listdir(self.data_path):
            paths.append(self.data_path)
            snomed_releases.append(self.release)
        else:
            for folder in os.listdir(self.data_path):
                if "SnomedCT" in folder:
                    paths.append(os.path.join(self.data_path, folder))
                    snomed_releases.append(folder[-16:-8])
        if len(paths) == 0:
            raise FileNotFoundError('Incorrect path to SNOMED CT directory')
        df2merge = []
        for i, snomed_release in enumerate(snomed_releases):
            refset_terminology = f'{paths[i]}/Snapshot/Refset/Map'
            snomed_v = ''
            opcs4_ref_set = 'der2_iisssccRefset_ExtendedMapSnapshot'
            if self.uk_ext:
                if "SnomedCT_InternationalRF2_PRODUCTION" in paths[i]:
                    continue
                elif "SnomedCT_UKClinicalRF2_PRODUCTION" in paths[i]:
                    opcs4_ref_set = "der2_iisssciRefset_ExtendedMapUKCLSnapshot"
                elif "SnomedCT_UKEditionRF2_PRODUCTION" in paths[i]:
                    opcs4_ref_set = "der2_iisssciRefset_ExtendedMapUKEDSnapshot"
                elif "SnomedCT_UKClinicalRefsetsRF2_PRODUCTION" in paths[i]:
                    continue
                else:
                    pass
            for f in os.listdir(refset_terminology):
                m = re.search(f'{opcs4_ref_set}'+r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)
            if snomed_v == '':
                raise FileNotFoundError("This SNOMED release does not contain OPCS mapping files")
            mappings = parse_file(f'{refset_terminology}/{opcs4_ref_set}_{snomed_v}_{snomed_release}.txt')
            mappings = mappings[mappings.active == '1']
            icd_mappings = mappings.sort_values(by=['referencedComponentId', 'mapPriority', 'mapGroup']).reset_index(
                drop=True)
            df2merge.append(icd_mappings)
        return pd.concat(df2merge)



import logging
import gc
from joblib import Parallel, delayed
from typing import Iterable, Generator, Tuple, Callable, Union, Iterator
from spacy.tokens import Doc, Span
from spacy.tokens.underscore import Underscore
from spacy.pipeline import Pipe
from spacy.util import minibatch


class PipeRunner(Pipe):

    log = logging.getLogger(__name__)
    _execute = None
    _delayed = None
    _time_out_in_secs = 3600

    def __init__(self, workers: int):
        self.workers = workers

    def __call__(self, doc: Doc):
        raise NotImplementedError("Method __call__ has not been implemented.")

    # Override
    def pipe(self, stream: Iterable[Doc], batch_size: int, **kwargs) -> Union[Generator[Doc, None, None], Iterator[Doc]]:
        error_handler = self.get_error_handler()
        if kwargs.get("parallel", False):
            PipeRunner._execute, PipeRunner._delayed = self._lazy_init_pool()
            for docs in minibatch(stream, size=self.workers):
                docs = [PipeRunner.serialize_entities(doc) for doc in docs]
                try:
                    tasks = (PipeRunner._delayed(self.__call__, doc, Underscore.get_state()) for doc in docs)
                    for output_doc in PipeRunner._execute(tasks):
                        yield PipeRunner.deserialize_entities(output_doc)
                except Exception as e:
                    error_handler(self.name, self, docs, e)
                    yield from [None] * len(docs)
        else:
            for doc in stream:
                try:
                    yield self(doc)
                except Exception as e:
                    error_handler(self.name, self, [doc], e)
                    yield None

    @staticmethod
    def serialize_entities(doc: Doc):
        new_ents = []
        for ent in doc._.ents:
            serializable = {
                "start": ent.start,
                "end": ent.end,
                "label": ent.label_,
                "cui": ent._.cui,
                "detected_name": ent._.detected_name,
                "context_similarity": ent._.context_similarity,
                "link_candidates": ent._.link_candidates,
                "confidence": ent._.confidence,
                "id": ent._.id
            }
            if hasattr(ent._, 'meta_anns') and ent._.meta_anns:
                serializable['meta_anns'] = ent._.meta_anns
            new_ents.append(serializable)
        doc._.ents.clear()
        gc.collect()
        doc._.ents = new_ents
        return doc

    @staticmethod
    def deserialize_entities(doc: Doc):
        new_ents = []
        for ent in doc._.ents:
            ent_span = Span(doc, ent['start'], ent['end'], label=ent['label'])
            ent_span._.cui = ent['cui']
            ent_span._.detected_name = ent['detected_name']
            ent_span._.context_similarity = ent['context_similarity']
            ent_span._.link_candidates = ent['link_candidates']
            ent_span._.confidence = ent['confidence']
            ent_span._.id = ent['id']
            if 'meta_anns' in ent:
                ent_span._.meta_anns = ent['meta_anns']
            new_ents.append(ent_span)
        doc._.ents.clear()
        gc.collect()
        doc._.ents = new_ents
        return doc

    @staticmethod
    def _run_pipe_on_one(call: Callable, doc: Doc, underscore_state: Tuple) -> Doc:
        Underscore.load_state(underscore_state)
        doc = PipeRunner.deserialize_entities(doc)
        doc = call(doc)
        doc = PipeRunner.serialize_entities(doc)
        return doc

    def _lazy_init_pool(self) -> Tuple:
        if PipeRunner._execute is None or self.workers > PipeRunner._execute.n_jobs:
            PipeRunner._execute = Parallel(n_jobs=self.workers, timeout=PipeRunner._time_out_in_secs)
        if PipeRunner._delayed is None:
            PipeRunner._delayed = delayed(PipeRunner._run_pipe_on_one)
        return PipeRunner._execute, PipeRunner._delayed



import xxhash
import dill
from io import BytesIO as StringIO


def dumps(obj, length=False):
    file = StringIO()
    dill.dump(obj, file, recurse=True)
    if length:
        return dumps(len(file.getvalue()), length=False)
    else:
        return file.getvalue()


class Hasher(object):
    def __init__(self):
        self.m = xxhash.xxh64()

    def update(self, obj, length=False):
        self.m.update(dumps(obj, length=length))

    def update_bytes(self, b):
        self.m.update(b)

    def hexdigest(self):
        return self.m.hexdigest()



import os
import logging
import time
from dataclasses import dataclass
from typing import List, Tuple, Optional, TypeVar, Type
from medcat.cdb import CDB
from medcat.utils.decorators import check_positive

T = TypeVar("T", bound="Checkpoint")


class Checkpoint(object):
    r""" The base class of checkpoint objects

    Args:
        dir_path (str):
            The path to the parent directory of checkpoint files
        steps (int):
            The number of processed sentences/documents before a checkpoint is saved
            (N.B.: A small number could result in error "no space left on device")
        max_to_keep (int):
            The maximum number of checkpoints to keep
            (N.B.: A large number could result in error "no space left on device")
    """
    DEFAULT_STEP = 1000
    DEFAULT_MAX_TO_KEEP = 1
    log = logging.getLogger(__package__)

    @check_positive
    def __init__(self, dir_path: str, *, steps: int = DEFAULT_STEP, max_to_keep: int = DEFAULT_MAX_TO_KEEP) -> None:
        self._dir_path = os.path.abspath(dir_path)
        self._steps = steps
        self._max_to_keep = max_to_keep
        self._file_paths: List[str] = []
        self._count = 0
        os.makedirs(self._dir_path, exist_ok=True)

    @property
    def steps(self) -> int:
        return self._steps

    @steps.setter
    def steps(self, value: int) -> None:
        check_positive(lambda _: ...)(value)    # [https://github.com/python/mypy/issues/1362]
        self._steps = value

    @property
    def max_to_keep(self) -> int:
        return self._max_to_keep

    @max_to_keep.setter
    def max_to_keep(self, value: int) -> None:
        check_positive(lambda _: ...)(value)    # [https://github.com/python/mypy/issues/1362]
        self._max_to_keep = value

    @property
    def count(self) -> int:
        return self._count

    @property
    def dir_path(self) -> str:
        return self._dir_path

    @classmethod
    def from_latest(cls: Type[T], dir_path: str) -> T:
        r'''
        Retrieve the latest checkpoint from the parent directory.

        Args:
            dir_path (string):
                The path to the directory containing checkpoint files
        Returns:
            A new checkpoint object
        '''
        if not os.path.isdir(dir_path):
            raise Exception("Checkpoints not found. You need to train from scratch.")
        ckpt_file_paths = cls._get_ckpt_file_paths(dir_path)
        if not ckpt_file_paths:
            raise Exception("Checkpoints not found. You need to train from scratch.")
        latest_ckpt = ckpt_file_paths[-1]
        steps, count = cls._get_steps_and_count(latest_ckpt)

        checkpoint = cls(dir_path, steps=steps)
        checkpoint._file_paths = ckpt_file_paths
        checkpoint._count = count
        cls.log.info(f"Checkpoint loaded from {latest_ckpt}")
        return checkpoint

    def save(self, cdb: CDB, count: int) -> None:
        r'''
        Save the CDB as the latest checkpoint.

        Args:
            cdb (medcat.CDB):
                The MedCAT CDB object to be checkpointed
            count (count):
                The number of the finished steps
        '''
        ckpt_file_path = os.path.join(os.path.abspath(self._dir_path), "checkpoint-%s-%s" % (self.steps, count))
        while len(self._file_paths) >= self._max_to_keep:
            to_remove = self._file_paths.pop(0)
            os.remove(to_remove)
        cdb.save(ckpt_file_path)
        self.log.debug("Checkpoint saved: %s", ckpt_file_path)
        self._file_paths.append(ckpt_file_path)
        self._count = count

    def restore_latest_cdb(self) -> CDB:
        r'''
        Restore the CDB from the latest checkpoint.

        Returns:
            cdb (medcat.CDB):
                The MedCAT CDB object
        '''
        if not os.path.isdir(self._dir_path):
            raise Exception("Checkpoints not found. You need to train from scratch.")
        ckpt_file_paths = self._get_ckpt_file_paths(self._dir_path)
        if not ckpt_file_paths:
            raise Exception("Checkpoints not found. You need to train from scratch.")
        latest_ckpt = ckpt_file_paths[-1]
        _, count = self._get_steps_and_count(latest_ckpt)
        self._file_paths = ckpt_file_paths
        self._count = count
        return CDB.load(self._file_paths[-1])

    @staticmethod
    def _get_ckpt_file_paths(dir_path: str) -> List[str]:
        ckpt_file_paths = [os.path.abspath(os.path.join(dir_path, f)) for f in os.listdir(dir_path)]
        ckpt_file_paths = [f for f in ckpt_file_paths if os.path.isfile(f) and "checkpoint-" in f]
        if ckpt_file_paths:
            ckpt_file_paths.sort(key=lambda f: Checkpoint._get_steps_and_count(f)[1])
        return ckpt_file_paths

    @staticmethod
    def _get_steps_and_count(file_path) -> Tuple[int, int]:
        file_name_parts = os.path.basename(file_path).split('-')
        return int(file_name_parts[1]), int(file_name_parts[2])


@dataclass
class CheckpointConfig(object):
    output_dir: str = "checkpoints"
    steps: int = Checkpoint.DEFAULT_STEP
    max_to_keep: int = Checkpoint.DEFAULT_MAX_TO_KEEP


class CheckpointManager(object):
    r"""
    The class for managing checkpoints of specific training type and their configuration

    Args:
        name (str):
            The name of the checkpoint manager (also used as the checkpoint base directory name).
        checkpoint_config (medcat.utils.checkpoint.CheckpointConfig):
            The checkpoint config object.
    """
    def __init__(self, name: str, checkpoint_config: CheckpointConfig) -> None:
        self.name = name
        self.checkpoint_config = checkpoint_config

    def create_checkpoint(self, dir_path: Optional[str] = None) -> "Checkpoint":
        r'''
        Create a new checkpoint inside the checkpoint base directory.

        Args:
            dir_path (str):
                The path to the checkpoint directory

        Returns:
            A checkpoint object
        '''
        dir_path = dir_path or os.path.join(os.path.abspath(os.getcwd()), self.checkpoint_config.output_dir, self.name, str(int(time.time())))
        return Checkpoint(dir_path,
                          steps=self.checkpoint_config.steps,
                          max_to_keep=self.checkpoint_config.max_to_keep)

    def get_latest_checkpoint(self, base_dir_path: Optional[str] = None) -> "Checkpoint":
        r'''
        Retrieve the latest checkpoint from the checkpoint base directory.

        Args:
            base_dir_path (string):
                The path to the directory containing checkpoint files

        Returns:
            A checkpoint object
        '''
        base_dir_path = base_dir_path or os.path.join(os.path.abspath(os.getcwd()), self.checkpoint_config.output_dir, self.name)
        ckpt_dir_path = self.get_latest_training_dir(base_dir_path=base_dir_path)
        checkpoint = Checkpoint.from_latest(dir_path=ckpt_dir_path)
        checkpoint.steps = self.checkpoint_config.steps
        checkpoint.max_to_keep = self.checkpoint_config.max_to_keep
        return checkpoint

    @classmethod
    def get_latest_training_dir(cls, base_dir_path: str) -> str:
        r'''
        Retrieve the latest training directory containing all checkpoints.

        Args:
            base_dir_path (string):
                The path to the directory containing all checkpointed trainings
        Returns:
            The path to the latest training directory containing all checkpoints.
        '''
        if not os.path.isdir(base_dir_path):
            raise ValueError(f"Checkpoint folder passed in does not exist: {base_dir_path}")
        ckpt_dir_paths = os.listdir(base_dir_path)
        if not ckpt_dir_paths:
            raise ValueError("No existing training found")
        ckpt_dir_paths.sort()
        ckpt_dir_path = os.path.abspath(os.path.join(base_dir_path, ckpt_dir_paths[-1]))
        return ckpt_dir_path



# Generated by Django 2.2.3 on 2019-09-17 11:43

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='UploadedText',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.TextField(blank=True, default='')),
                ('create_time', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]



import unittest
import pickle
import tempfile
from medcat.config import Config


class ConfigTests(unittest.TestCase):

    def test_pickleability(self):
        with tempfile.TemporaryFile() as f:
            pickle.dump(Config(), f)

    def test_from_dict(self):
        config = Config.from_dict({"key": "value"})
        self.assertEqual("value", config.key)


if __name__ == '__main__':
    unittest.main()



def check_filters(cui, filters):
    r''' Checks is a CUI in the filters

    Args:
        cui
        filters

    Return:
        bool:
            True if in filters else False
    '''
    if cui in filters.get('cuis', {}) or not filters.get('cuis', {}):
        return cui not in filters.get('cuis_exclude', {})
    else:
        return False


def get_all_irrelevant_cuis(project, cdb):
    i_cuis = set()
    for d in project['documents']:
        for a in d['annotations']:
            if 'irrelevant' in a and a['irrelevant']:
                i_cuis.add(a['cui'])
    return i_cuis


def get_project_filters(cuis, type_ids, cdb, project=None):
    cui_filter = set()
    if isinstance(cuis, str):
        if cuis is not None and cuis:
            cui_filter.update([x.strip() for x in cuis.split(",")])
        if type_ids is not None and type_ids:
            type_ids = [x.strip().upper() for x in type_ids.split(",")]

            # Convert type_ids to cuis
            if 'type_id2cuis' in cdb.addl_info:
                for type_id in type_ids:
                    if type_id in cdb.addl_info['type_id2cuis']:
                        cui_filter.update(cdb.addl_info['type_id2cuis'][type_id])
                    else:
                        raise Exception("Impossible to create filters, disable them.")
            else:
                raise Exception("Impossible to create filters, disable them.")
    elif isinstance(cuis, list):
        cui_filter = set(cuis)

    if project is not None:
        i_cuis = get_all_irrelevant_cuis(project, cdb)
        print(i_cuis)
        for i_cui in i_cuis:
            cui_filter.remove(i_cui)

    return cui_filter



import logging
import os
import unittest
import numpy as np
from timeit import default_timer as timer
from medcat.cdb import CDB
from medcat.preprocessing.tokenizers import spacy_split_all
from medcat.ner.vocab_based_ner import NER
from medcat.preprocessing.taggers import tag_skip_and_punct
from medcat.pipe import Pipe
from medcat.utils.normalizers import BasicSpellChecker
from medcat.vocab import Vocab
from medcat.preprocessing.cleaners import prepare_name
from medcat.linking.vector_context_model import ContextModel
from medcat.linking.context_based_linker import Linker
from medcat.config import Config


class NerArchiveTests(unittest.TestCase):

    def setUp(self) -> None:
        self.config = Config()
        self.config.general['log_level'] = logging.INFO
        cdb = CDB(config=self.config)

        self.nlp = Pipe(tokenizer=spacy_split_all, config=self.config)
        self.nlp.add_tagger(tagger=tag_skip_and_punct,
                       name='skip_and_punct',
                       additional_fields=['is_punct'])

        # Add a couple of names
        cdb.add_names(cui='S-229004', names=prepare_name('Movar', self.nlp, {}, self.config))
        cdb.add_names(cui='S-229004', names=prepare_name('Movar viruses', self.nlp, {}, self.config))
        cdb.add_names(cui='S-229005', names=prepare_name('CDB', self.nlp, {}, self.config))
        # Check
        #assert cdb.cui2names == {'S-229004': {'movar', 'movarvirus', 'movarviruses'}, 'S-229005': {'cdb'}}

        self.vocab_path = "./tmp_vocab.dat"
        if not os.path.exists(self.vocab_path):
            import requests
            tmp = requests.get("https://medcat.rosalind.kcl.ac.uk/media/vocab.dat")
            with open(self.vocab_path, 'wb') as f:
                f.write(tmp.content)

        vocab = Vocab.load(self.vocab_path)
        # Make the pipeline
        self.nlp = Pipe(tokenizer=spacy_split_all, config=self.config)
        self.nlp.add_tagger(tagger=tag_skip_and_punct,
                       name='skip_and_punct',
                       additional_fields=['is_punct'])
        spell_checker = BasicSpellChecker(cdb_vocab=cdb.vocab, config=self.config, data_vocab=vocab)
        self.nlp.add_token_normalizer(spell_checker=spell_checker, config=self.config)
        ner = NER(cdb, self.config)
        self.nlp.add_ner(ner)

        # Add Linker
        link = Linker(cdb, vocab, self.config)
        self.nlp.add_linker(link)

        self.text = "CDB - I was running and then Movar    Virus attacked and CDb"

    def tearDown(self) -> None:
        self.nlp.destroy()

    def test_limits_for_tokens_and_uppercase(self):
        self.config.ner['max_skip_tokens'] = 1
        self.config.ner['upper_case_limit_len'] = 4
        self.config.linking['disamb_length_limit'] = 2

        d = self.nlp(self.text)

        assert len(d._.ents) == 2
        assert d._.ents[0]._.link_candidates[0] == 'S-229004'

    def test_change_limit_for_skip(self):
        self.config.ner['max_skip_tokens'] = 3
        d = self.nlp(self.text)
        assert len(d._.ents) == 3

    def test_change_limit_for_upper_case(self):
        self.config.ner['upper_case_limit_len'] = 3
        d = self.nlp(self.text)
        assert len(d._.ents) == 4

    def test_check_name_length_limit(self):
        self.config.ner['min_name_len'] = 4
        d = self.nlp(self.text)
        assert len(d._.ents) == 2

    def test_speed(self):
        text = "CDB - I was running and then Movar    Virus attacked and CDb"
        text = text * 300
        self.config.general['spell_check'] = True
        start = timer()
        for i in range(50):
            d = self.nlp(text)
        end = timer()
        print("Time: ", end - start)

    def test_without_spell_check(self):
        # Now without spell check
        self.config.general['spell_check'] = False
        start = timer()
        for i in range(50):
            d = self.nlp(self.text)
        end = timer()
        print("Time: ", end - start)


    def test_for_linker(self):
        self.config = Config()
        self.config.general['log_level'] = logging.DEBUG
        cdb = CDB(config=self.config)

        # Add a couple of names
        cdb.add_names(cui='S-229004', names=prepare_name('Movar', self.nlp, {}, self.config))
        cdb.add_names(cui='S-229004', names=prepare_name('Movar viruses', self.nlp, {}, self.config))
        cdb.add_names(cui='S-229005', names=prepare_name('CDB', self.nlp, {}, self.config))
        cdb.add_names(cui='S-2290045', names=prepare_name('Movar', self.nlp, {}, self.config))
        # Check
        #assert cdb.cui2names == {'S-229004': {'movar', 'movarvirus', 'movarviruses'}, 'S-229005': {'cdb'}, 'S-2290045': {'movar'}}

        cuis = list(cdb.cui2names.keys())
        for cui in cuis[0:50]:
            vectors = {'short': np.random.rand(300),
                      'long': np.random.rand(300),
                      'medium': np.random.rand(300)
                      }
            cdb.update_context_vector(cui, vectors, negative=False)

        d = self.nlp(self.text)
        vocab = Vocab.load(self.vocab_path)
        cm = ContextModel(cdb, vocab, self.config)
        cm.train_using_negative_sampling('S-229004')
        self.config.linking['train_count_threshold'] = 0

        cm.train('S-229004', d._.ents[1], d)

        cm.similarity('S-229004', d._.ents[1], d)

        cm.disambiguate(['S-2290045', 'S-229004'], d._.ents[1], 'movar', d)



import pandas as pd
from medcat.config import Config
from medcat.cat import CAT
from medcat.cdb_maker import CDBMaker


class RepairCDB(object):
    def __init__(self, base_cdb, final_cdb, vocab):
        self.base_cdb = base_cdb
        self.vocab = vocab
        self.final_cdb = final_cdb
        self.final_cat = None
        self.cdb = None
        self.cat = None
        self.base_cat = None

    def prepare(self, cuis):
        self.base_cdb.filter_by_cui(cuis)

        csv = [['cui', 'name']]
        names = set()
        cui = 0
        for base_cui in self.base_cdb.cui2names:
            if self.base_cdb.cui2context_vectors.get(base_cui, {}):
                for name in self.base_cdb.cui2names[base_cui]:
                    if name not in names and name in self.base_cdb.name2cuis:
                        csv.append([cui, name.replace("~", " ")])
                        cui += 1
                        names.add(name)

        df_cdb = pd.DataFrame(csv[1:], columns=csv[0])
        df_cdb.to_csv("/tmp/data.csv", index=False)

        config = Config()
        cdb_maker = CDBMaker(config=config)

        cdb = cdb_maker.prepare_csvs(['/tmp/data.csv'])

        # Rempove ambigous
        for name in cdb.name2cuis:
            cuis = cdb.name2cuis[name]
            if len(cuis) > 1:
                cnts = []
                for cui in cuis:
                    cnts.append([cui, len(cdb.cui2names[cui])])
                cnts.sort(key=lambda x: x[1])
                cdb.name2cuis[name] = [cnts[-1][0]]

        self.cdb = cdb
        self.base_cdb.reset_cui_count(n=10)
        self.cat = CAT(cdb=self.cdb, config=self.cdb.config, vocab=self.vocab)
        self.base_cat = CAT(cdb=self.base_cdb, config=self.base_cdb.config, vocab=self.vocab)

    def train(self, data_iterator, n_docs=100000):
        docs = []
        for doc in data_iterator:
            docs.append(doc)
            if len(docs) >= n_docs:
                break
        self.cat.train(data_iterator=docs)
        self.base_cat.train(data_iterator=docs)

    def calculate_scores(self, count_limit=1000):
        data = [['new_cui', 'base_cui', 'name', 'new_count', 'base_count', 'score', 'decision']]
        for name, cuis2 in self.cdb.name2cuis.items():
            cui2 = cuis2[0]
            count2 = self.cdb.cui2count_train.get(cui2, 0)
            if count2 > count_limit:
                cuis = self.base_cdb.name2cuis.get(name, [])
                for cui in cuis:
                    count = self.base_cdb.cui2count_train.get(cui, 0)
                    if self.base_cdb.cui2context_vectors.get(cui, {}):
                        score = count2 / count
                        data.append([cui2, cui, name, count2, count, score, ''])

        self.scores_df = pd.DataFrame(data[1:], columns=data[0])


    def unlink_names(self, sort='score', skip=0, cui_filter=None, apply_existing_decisions=0):
        scores_df = self.scores_df.sort_values(sort, ascending=False)
        self.final_cdb.config.general['full_unlink'] = False
        if self.final_cat is None:
            self.final_cat = CAT(cdb=self.final_cdb, config=self.final_cdb.config, vocab=self.vocab)

        for ind, row in enumerate(scores_df.iterrows()):
            row_ind, row = row
            if ind < skip:
                continue
            name = row['name']
            base_cui = row['base_cui']
            new_cui = row['new_cui']
            base_count = row['base_count']
            new_count = row['new_count']
            cui = row['base_cui']
            if base_cui in cui_filter:
                print("{:3} -- {:20} -> {:20}, base_count: {}, new_count: {}, cui: {}".format(
                    ind, str(name)[:20], str(self.final_cdb.get_name(base_cui))[:30], base_count, new_count, cui))

                if apply_existing_decisions and apply_existing_decisions > ind:
                    decision = row['decision']
                else:
                    decision = input("Decision (l/...): ")

                if decision == 'l':
                    names = self.cdb.cui2names[new_cui]
                    print("Unlinking: " + str(names))
                    print("\n\n")
                    for name in names:
                        self.final_cat.unlink_concept_name(base_cui, name, preprocessed_name=True)
                elif decision == 'f':
                    if base_cui in cui_filter:
                        print("Removing from filter: " + str(base_cui))
                        print("\n\n")
                        cui_filter.remove(base_cui)
                else:
                    decision = 'k' # Means keep

                self.scores_df.iat[row_ind, 6] = decision



"""Loggers
"""
import logging


def add_handlers(log):
    """ The most basic logger type, log is just:
    logger_name: <basic_info>

    name:  logger name
    """
    if len(log.handlers) == 0: # If we do not have any handlers add them
        # create a file handler
        fh = logging.FileHandler('medcat.log')
        # create console handler
        ch = logging.StreamHandler()

        # add the handlers to the logger
        log.addHandler(fh)
        log.addHandler(ch)

    return log



import os
import shutil
import unittest
import tempfile
import asyncio
from medcat.config import Config
from medcat.cdb_maker import CDBMaker


class CDBTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        config = Config()
        config.general["spacy_model"] = "en_core_web_md"
        cls.cdb_maker = CDBMaker(config)

    def setUp(self) -> None:
        cdb_csv = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "examples", "cdb.csv")
        cdb_2_csv = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "examples", "cdb_2.csv")
        self.tmp_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "tmp")
        os.makedirs(self.tmp_dir, exist_ok=True)
        self.undertest = CDBTests.cdb_maker.prepare_csvs([cdb_csv, cdb_2_csv], full_build=True)

    def tearDown(self) -> None:
        shutil.rmtree(self.tmp_dir)

    def test_name2cuis(self):
        self.assertEqual({
            'second~csv': ['C0000239'],
            'virus': ['C0000039', 'C0000139'],
            'virus~k': ['C0000039', 'C0000139'],
            'virus~m': ['C0000039', 'C0000139'],
            'virus~z': ['C0000039', 'C0000139']
        }, self.undertest.name2cuis)

    def test_cui2names(self):
        self.assertEqual({
            'C0000039': {'virus~z', 'virus~k', 'virus~m', 'virus'},
            'C0000139': {'virus~z', 'virus', 'virus~m', 'virus~k'},
            'C0000239': {'second~csv'}
        }, self.undertest.cui2names)

    def test_cui2preferred_name(self):
        self.assertEqual({'C0000039': 'Virus', 'C0000139': 'Virus Z'}, self.undertest.cui2preferred_name)

    def test_cui2type_ids(self):
        self.assertEqual({'C0000039': {'T109', 'T234', 'T123'}, 'C0000139': set(), 'C0000239': set()}, self.undertest.cui2type_ids)

    def test_save_and_load(self):
        with tempfile.NamedTemporaryFile() as f:
            self.undertest.save(f.name)
            self.undertest.load(f.name)

    def test_save_async_and_load(self):
        with tempfile.NamedTemporaryFile() as f:
            asyncio.run(self.undertest.save_async(f.name))
            self.undertest.load(f.name)


if __name__ == '__main__':
    unittest.main()



from email.policy import default
from django import forms
from .models import Downloader


class DownloaderForm(forms.ModelForm):
    consent = forms.BooleanField(required=True, label=(
        f"I consent to MedCAT collecting and storing my names, email, company"
        f" or academic institution name, funder and project title, and use"
        f" case description. I am aware that MedCAT has been funded through"
        f" academic research grants, and therefore funding bodies require its"
        f" support team to report wider impact and usage of produced works"
        f" with the above information."
    ))

    def __init__(self, models, *args, **kwargs):
        super().__init__(*args, *kwargs)
        self.fields["modelpack"] = forms.ChoiceField(label="Select a model for download",
                                                     choices=[(model.model_name, model.model_display_name) for model in models],
                                                     widget=forms.RadioSelect())

    class Meta:
        model = Downloader
        fields = [
            "first_name",
            "last_name",
            "email",
            "affiliation",
            "funder",
            "use_case",
        ]
        labels = {
            "first_name": "First Name",
            "last_name": "Last Name",
            "email": "Email",
            "affiliation": "Company or Academic Institution",
            "funder": "Funder and Project Title (optional)",
            "use_case": "Please describe your use case",
        }
        widgets = {
            "affiliation": forms.TextInput(attrs={"size": 40}),
            "funder": forms.TextInput(attrs={"size": 40}),
            "use_case": forms.Textarea(attrs={"rows": 5, "cols": 40}),
        }



from django.db import models
from django.core.files.storage import FileSystemStorage


MODEL_FS = FileSystemStorage(location="/medcat_data")


# Create your models here.
class UploadedText(models.Model):
    text = models.TextField(default="", blank=True)
    create_time = models.DateTimeField(auto_now_add=True)


class Downloader(models.Model):
    first_name = models.CharField(max_length=20)
    last_name = models.CharField(max_length=20)
    email = models.EmailField(max_length=50)
    affiliation = models.CharField(max_length=100)
    funder = models.CharField(max_length=100, blank=True, default="")
    use_case = models.TextField(max_length=200)

    def __str__(self):
        return f'{self.first_name} - {self.last_name}'


class MedcatModel(models.Model):
    model_name = models.CharField(max_length=20, unique=True)
    model_file = models.FileField(storage=MODEL_FS)
    model_display_name = models.CharField(max_length=50)



import torch
from collections import OrderedDict
from typing import Optional, Any, List
from torch import nn, Tensor
from torch.nn import CrossEntropyLoss
from transformers import BertPreTrainedModel, BertModel, BertConfig
from transformers.modeling_outputs import TokenClassifierOutput
from medcat.meta_cat import ConfigMetaCAT


class LSTM(nn.Module):
    def __init__(self, embeddings: Optional[Tensor], config: ConfigMetaCAT) -> None:
        super(LSTM, self).__init__()

        self.config = config
        # Get the required sizes
        vocab_size = config.general['vocab_size']
        embedding_size = config.model['input_size']

        # Initialize embeddings
        self.embeddings = nn.Embedding(vocab_size, embedding_size, padding_idx=config.model['padding_idx'])
        if embeddings is not None:
            self.embeddings.load_state_dict(OrderedDict([('weight', embeddings)]))
            # Disable training for the embeddings - IMPORTANT
            self.embeddings.weight.requires_grad = config.model['emb_grad']

        # Create the RNN cell - devide 
        self.rnn = nn.LSTM(input_size=config.model['input_size'],
                           hidden_size=config.model['hidden_size'] // config.model['num_directions'],
                           num_layers=config.model['num_layers'],
                           dropout=config.model['dropout'],
                           bidirectional=config.model['num_directions'] == 2)
        self.fc1 = nn.Linear(config.model['hidden_size'], config.model['nclasses'])

        self.d1 = nn.Dropout(config.model['dropout'])

    def forward(self,
                input_ids: torch.LongTensor,
                center_positions: Tensor,
                attention_mask: Optional[torch.FloatTensor] = None,
                ignore_cpos: bool = False) -> Tensor:
        x = input_ids
        # Get the mask from x
        if attention_mask is None:
            mask = x != self.config.model['padding_idx']
        else:
            mask = attention_mask

        # Embed the input: from id -> vec
        x = self.embeddings(x) # x.shape = batch_size x sequence_length x emb_size

        # Tell RNN to ignore padding and set the batch_first to True
        x = nn.utils.rnn.pack_padded_sequence(x, mask.sum(1).int().view(-1).cpu(), batch_first=True, enforce_sorted=False)

        # Run 'x' through the RNN
        x, hidden = self.rnn(x)

        # Add the padding again
        x, _ = torch.nn.utils.rnn.pad_packed_sequence(x, batch_first=True)

        # Get what we need
        row_indices = torch.arange(0, x.size(0)).long()

        # If this is  True we will always take the last state and not CPOS
        if ignore_cpos:
            x = hidden[0]
            x = x.view(self.config.model['num_layers'], self.config.model['num_directions'], -1,
                       self.config.model['hidden_size']//self.config.model['num_directions'])
            x = x[-1, :, :, :].permute(1, 2, 0).reshape(-1, self.config.model['hidden_size'])
        else:
            x = x[row_indices, center_positions, :]

        # Push x through the fc network and add dropout
        x = self.d1(x)
        x = self.fc1(x)

        return x


class BertForMetaAnnotation(BertPreTrainedModel):

    _keys_to_ignore_on_load_unexpected: List[str] = [r"pooler"]  # type: ignore

    def __init__(self, config: BertConfig) -> None:
        super().__init__(config)
        self.num_labels = config.num_labels

        self.bert = BertModel(config, add_pooling_layer=False)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        self.init_weights()

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        center_positions: Optional[Any] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> TokenClassifierOutput:
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Labels for computing the token classification loss. Indices should be in ``[0, ..., config.num_labels -
            1]``.
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        sequence_output = outputs[0] # (batch_size, sequence_length, hidden_size)

        row_indices = torch.arange(0, sequence_output.size(0)).long()
        sequence_output = sequence_output[row_indices, center_positions, :]

        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)

        loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss()
            # Only keep active parts of the loss
            loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )



import os
import random
import math
import torch
import numpy as np
import torch.optim as optim
from typing import List, Optional, Tuple, Any, Dict
from torch import nn
from scipy.special import softmax
from medcat.config_meta_cat import ConfigMetaCAT
from medcat.tokenizers.meta_cat_tokenizers import TokenizerWrapperBase
from sklearn.metrics import classification_report, f1_score, precision_recall_fscore_support


def set_all_seeds(seed: int) -> None:
    torch.manual_seed(seed)
    np.random.seed(seed)


def create_batch_piped_data(data: List, start_ind: int, end_ind: int, device: torch.device, pad_id: int) -> Tuple:
    r''' Creates a batch given data and start/end that denote batch size, will also add
    padding and move to the right device.

    Args:
        data (List[List[int], int, Optional[int]]):
            Data in the format: [[<[input_ids]>, <cpos>, Optional[int]], ...], the third column is optional
            and represents the output label
        start_ind (`int`):
            Start index of this batch
        end_ind (`int`):
            End index of this batch
        device (`torch.device`):
            Where to move the data
        pad_id (`int`):
            Padding index

    Returns:
        x ():
            Same as data, but subsetted and as a tensor
        cpos ():
            Center positions for the data

    '''
    max_seq_len = max([len(x[0]) for x in data])
    x = [x[0][0:max_seq_len] + [pad_id]*max(0, max_seq_len - len(x[0])) for x in data[start_ind:end_ind]]
    cpos = [x[1] for x in data[start_ind:end_ind]]
    y = None
    if len(data[0]) == 3:
        # Means we have the y column
        y = torch.tensor([x[2] for x in data[start_ind:end_ind]], dtype=torch.long).to(device)

    x = torch.tensor(x, dtype=torch.long).to(device)
    cpos = torch.tensor(cpos, dtype=torch.long).to(device)

    return x, cpos, y


def predict(model: nn.Module, data: List, config: ConfigMetaCAT) -> Tuple:
    r''' Predict on data used in the meta_cat.pipe

    Args:
        data (List[List[List[int], int]]):
            Data in the format: [[<input_ids>, <cpos>], ...]
        config (medcat.config_meta_cat.ConfigMetaCAT):
            Configuration for this meta_cat instance.

    Returns:
        predictions (List[int]):
            For each row of input data a prediction
        confidence (List[float]):
            For each prediction a confidence value
    '''

    pad_id = config.model['padding_idx']
    batch_size = config.general['batch_size_eval']
    device = config.general['device']
    ignore_cpos = config.model['ignore_cpos']

    model.eval()
    model.to(device)

    num_batches = math.ceil(len(data) / batch_size)
    all_logits = []

    with torch.no_grad():
        for i in range(num_batches):
            x, cpos, _ = create_batch_piped_data(data, i*batch_size, (i+1)*batch_size, device=device, pad_id=pad_id)
            logits = model(x, cpos, ignore_cpos=ignore_cpos)
            all_logits.append(logits.detach().cpu().numpy())

    predictions = []
    confidences = []

    # Can be that there are not logits, data is empty
    if all_logits:
        logits = np.concatenate(all_logits, axis=0)
        predictions = np.argmax(logits, axis=1)
        confidences = np.max(softmax(logits, axis=1), axis=1)

    return predictions, confidences


def split_list_train_test(data: List, test_size: int, shuffle: bool = True) -> Tuple:
    r''' Shuffle and randomply split data

    Args:
        data
        test_size
        shuffle
    '''
    if shuffle:
        random.shuffle(data)

    test_ind = int(len(data) * test_size)
    test_data = data[:test_ind]
    train_data = data[test_ind:]

    return train_data, test_data


def print_report(epoch: int, running_loss: List, all_logits: List, y: Any, name: str = 'Train') -> None:
    r''' Prints some basic stats during training

    Args:
        epoch
        running_loss
        all_logits
        y
        name
    '''
    if all_logits:
        print(f'Epoch: {epoch} ' + "*"*50 + f"  {name}")
        print(classification_report(y, np.argmax(np.concatenate(all_logits, axis=0), axis=1)))


def train_model(model: nn.Module, data: List, config: ConfigMetaCAT, save_dir_path: Optional[str] = None) -> Dict:
    r''' Trains a LSTM model (for now) with autocheckpoints

    Args:
        data
        config
        save_dir_path
    '''
    # Get train/test from data
    train_data, test_data = split_list_train_test(data, test_size=config.train['test_size'], shuffle=config.train['shuffle_data'])
    device = torch.device(config.general['device']) # Create a torch device

    class_weights = config.train['class_weights']
    if class_weights is not None:
        class_weights = torch.FloatTensor(class_weights).to(device)
        criterion = nn.CrossEntropyLoss(weight=class_weights) # Set the criterion to Cross Entropy Loss
    else:
        criterion = nn.CrossEntropyLoss() # Set the criterion to Cross Entropy Loss
    parameters = filter(lambda p: p.requires_grad, model.parameters())
    optimizer = optim.Adam(parameters, lr=config.train['lr'])
    model.to(device) # Move the model to device

    batch_size = config.train['batch_size']
    batch_size_eval = config.general['batch_size_eval']
    pad_id = config.model['padding_idx']
    nepochs = config.train['nepochs']
    ignore_cpos = config.model['ignore_cpos']
    num_batches = math.ceil(len(train_data) / batch_size)
    num_batches_test = math.ceil(len(test_data) / batch_size_eval)

    # Can be pre-calculated for the whole dataset
    y_test = [x[2] for x in test_data]
    y_train = [x[2] for x in train_data]

    winner_report= {'f1': 0, 'report': '', 'epoch': 0}
    for epoch in range(nepochs):
        running_loss = []
        all_logits = []
        model.train()
        for i in range(num_batches):
            x, cpos, y = create_batch_piped_data(train_data, i*batch_size, (i+1)*batch_size, device=device, pad_id=pad_id)
            logits = model(x, center_positions=cpos, ignore_cpos=ignore_cpos)
            loss = criterion(logits, y)
            loss.backward()
            # Track loss and logits
            running_loss.append(loss.item())
            all_logits.append(logits.detach().cpu().numpy())

            parameters = filter(lambda p: p.requires_grad, model.parameters())
            nn.utils.clip_grad_norm_(parameters, 0.25)
            optimizer.step()

        all_logits_test = []
        running_loss_test = []
        model.eval()

        with torch.no_grad():
            for i in range(num_batches_test):
                x, cpos, y = create_batch_piped_data(test_data, i*batch_size_eval, (i+1)*batch_size_eval, device=device, pad_id=pad_id)
                logits = model(x, cpos, ignore_cpos=ignore_cpos)

                # Track loss and logits
                running_loss_test.append(loss.item())
                all_logits_test.append(logits.detach().cpu().numpy())

        print_report(epoch, running_loss, all_logits, y=y_train, name='Train')
        print_report(epoch, running_loss_test, all_logits_test, y=y_test, name='Test')

        score_average = config.train['score_average']
        f1 = f1_score(y_test, np.argmax(np.concatenate(all_logits_test, axis=0), axis=1), average=score_average)
        if f1 > winner_report['f1']:
            report = classification_report(y_test, np.argmax(np.concatenate(all_logits_test, axis=0), axis=1), output_dict=True)
            winner_report['f1'] = f1
            winner_report['report'] = report
            winner_report['epoch'] = epoch

            # Save if needed
            if config.train['auto_save_model']:
                if save_dir_path is None:
                    raise Exception(
                        "The `save_dir_path` argument is required if `aut_save_model` is `True` in the config")
                else:
                    path = os.path.join(save_dir_path, 'model.dat')
                    torch.save(model.state_dict(), path)
                    print("\n##### Model saved to {} at epoch: {} and f1: {} #####\n".format(path, epoch, f1))

    return winner_report


def eval_model(model: nn.Module, data: List, config: ConfigMetaCAT, tokenizer: TokenizerWrapperBase) -> Dict:
    r''' Evaluate a trained model on the provided data

    Args:
        model
        data
        config

    '''
    device = torch.device(config.general['device']) # Create a torch device
    batch_size_eval = config.general['batch_size_eval']
    pad_id = config.model['padding_idx']
    ignore_cpos = config.model['ignore_cpos']
    class_weights = config.train['class_weights']

    if class_weights is not None:
        class_weights = torch.FloatTensor(class_weights).to(device)
        criterion = nn.CrossEntropyLoss(weight=class_weights) # Set the criterion to Cross Entropy Loss
    else:
        criterion = nn.CrossEntropyLoss() # Set the criterion to Cross Entropy Loss

    y_eval = [x[2] for x in data]
    num_batches = math.ceil(len(data) / batch_size_eval)
    running_loss = []
    all_logits = []
    model.to(device)
    model.eval()

    with torch.no_grad():
        for i in range(num_batches):
            x, cpos, y = create_batch_piped_data(data, i*batch_size_eval, (i+1)*batch_size_eval, device=device, pad_id=pad_id)
            logits = model(x, cpos, ignore_cpos=ignore_cpos)
            loss = criterion(logits, y)

            # Track loss and logits
            running_loss.append(loss.item())
            all_logits.append(logits.detach().cpu().numpy())

    print_report(0, running_loss, all_logits, y=y_eval, name='Eval')

    score_average = config.train['score_average']
    predictions = np.argmax(np.concatenate(all_logits, axis=0), axis=1)
    precision, recall, f1, support = precision_recall_fscore_support(y_eval, predictions, average=score_average)

    examples: Dict = {'FP': {}, 'FN': {}, 'TP': {}}
    id2category_value = {v: k for k, v in config.general['category_value2id'].items()}
    for i, p in enumerate(predictions):
        y = id2category_value[y_eval[i]]
        p = id2category_value[p]
        c = data[i][1]
        tkns = data[i][0]
        assert tokenizer.hf_tokenizers is not None
        text = tokenizer.hf_tokenizers.decode(tkns[0:c]) + " <<"+ tokenizer.hf_tokenizers.decode(tkns[c:c+1]).strip() + ">> " + \
            tokenizer.hf_tokenizers.decode(tkns[c+1:])
        info = "Predicted: {}, True: {}".format(p, y)
        if p != y:
            # We made a mistake
            examples['FN'][y] = examples['FN'].get(y, []) + [(info, text)]
            examples['FP'][p] = examples['FP'].get(p, []) + [(info, text)]
        else:
            examples['TP'][y] = examples['TP'].get(y, []) + [(info, text)]

    return {'precision': precision, 'recall': recall, 'f1': f1, 'examples': examples}



import unittest
from medcat.utils.decorators import check_positive


class DecoratorsTest(unittest.TestCase):

    def test_check_positive(self):
        @check_positive
        def func(value1, value2=None):
            return value1, value2

        with self.assertRaises(ValueError) as e:
            func(-1)
        self.assertEqual("Argument at position 0 is not a positive integer", str(e.exception))

        with self.assertRaises(ValueError) as e:
            func(1, -1)
        self.assertEqual("Argument at position 1 is not a positive integer", str(e.exception))



import os
from django.core import management
from django.conf import settings
from django_cron import CronJobBase, Schedule


class DbBackup(CronJobBase):

    RUN_EVERY_MINS = int(os.environ.get("DB_BACKUP_EVERY_MINS", 60 * 12))
    RETRY_AFTER_FAILURE_MINS = int(os.environ.get("DB_BACKUP_RETRY_BACKOFF_MINS", 5))

    schedule = Schedule(run_every_mins=RUN_EVERY_MINS, retry_after_failure_mins=RETRY_AFTER_FAILURE_MINS)
    code = "demo.db_backup.DbBackup"

    def __init__(self):
        backup_location = settings.DBBACKUP_STORAGE_OPTIONS["location"]
        os.makedirs(backup_location, exist_ok=True)

    def do(self):
        management.call_command("dbbackup", "--noinput", "-z")



from typing import Dict, Any
from medcat.config import ConfigMixin


class ConfigMetaCAT(ConfigMixin):

    def __init__(self) -> None:

        self.general: Dict[str, Any] = {
                'device': 'cpu',
                'disable_component_lock': False,
                'seed': 13,
                'description': "No description", # Should provide a basic description of this MetaCAT model
                'category_name': None, # What category is this meta_cat model predicting/training
                'category_value2id': {}, # Map from category values to ID, if empty it will be autocalculated during training
                'vocab_size': None, # Will be set automatically if the tokenizer is provided during meta_cat init
                'lowercase': True, # If true all input text will be lowercased
                'cntx_left': 15, # Number of tokens to take from the left of the concept
                'cntx_right': 10, # Number of tokens to take from the right of the concept
                'replace_center': None, # If set the center (concept) will be replaced with this string
                'batch_size_eval': 5000, # Number of annotations to be meta-annotated at once in eval
                'annotate_overlapping': False, # If set meta_anns will be calcualted for doc._.ents, otherwise for doc.ents
                'tokenizer_name': 'bbpe', # Tokenizer name used with of MetaCAT
                # This is a dangerous option, if not sure ALWAYS set to False. If set, it will try to share the pre-calculated
                #context tokens between MetaCAT models when serving. It will ignore differences in tokenizer and context size,
                #so you need to be sure that the models for which this is turned on have the same tokenizer and context size, during
                #a deployment.
                'save_and_reuse_tokens': False,
                'pipe_batch_size_in_chars': 20000000, # How many characters are piped at once into the meta_cat class
                }
        self.model: Dict[str, Any] = {
                'model_name': 'lstm',
                'num_layers': 2,
                'input_size': 300,
                'hidden_size': 300,
                'dropout': 0.5,
                'num_directions': 2, # 2 - bidirectional model, 1 - unidirectional
                'nclasses': 2, # Number of classes that this model will output
                'padding_idx': -1,
                'emb_grad': True, # If True the embeddings will also be trained
                'ignore_cpos': False, # If set to True center positions will be ignored when calculating represenation
                }

        self.train: Dict[str, Any] = {
                'batch_size': 100,
                'nepochs': 50,
                'lr': 0.001,
                'test_size': 0.1,
                'shuffle_data': True, # Used only during training, if set the dataset will be shuffled before train/test split
                'class_weights': None,
                'score_average': 'weighted', # What to use for averaging F1/P/R across labels
                'prerequisites': {},
                'cui_filter': None, # If set only this CUIs will be used for training
                'auto_save_model': True, # Should do model be saved during training for best results
                'last_train_on': None, # When was the last training run
                }



import unittest
from spacy.lang.en import English
from spacy.tokens import Span
from medcat.pipeline.pipe_runner import PipeRunner


class PipeRunnerTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.text = "CDB - I was running and then Movar Virus attacked and CDb"
        cls.nlp = English()

    def test_pipe_multi_workers(self):
        docs = list(_PipeRunnerImpl(workers=2).pipe(
            [self.nlp.make_doc(self.text), self.nlp.make_doc(self.text), self.nlp.make_doc(self.text)],
            batch_size=1,
            parallel=True
        ))

        self.assertEqual(3, len(docs))
        self.assertEqual(self.text, docs[0].text)
        self.assertEqual(self.text, docs[1].text)
        self.assertEqual(self.text, docs[2].text)

    def test_pipe_single_worker(self):
        docs = list(_PipeRunnerImpl(workers=1).pipe(
            [self.nlp.make_doc(self.text), self.nlp.make_doc(self.text), self.nlp.make_doc(self.text)],
            batch_size=1,
            parallel=True
        ))

        self.assertEqual(3, len(docs))
        self.assertEqual(self.text, docs[0].text)
        self.assertEqual(self.text, docs[1].text)
        self.assertEqual(self.text, docs[2].text)

    def test_pipe_no_workers(self):
        docs = list(_PipeRunnerImpl(workers=2).pipe(
            [self.nlp.make_doc(self.text), self.nlp.make_doc(self.text), self.nlp.make_doc(self.text)],
            batch_size=1,
            parallel=False
        ))

        self.assertEqual(3, len(docs))
        self.assertEqual(self.text, docs[0].text)
        self.assertEqual(self.text, docs[1].text)
        self.assertEqual(self.text, docs[2].text)

    def test_serialize_entities(self):
        doc = self.nlp.make_doc(self.text)
        doc._.ents = [Span(doc, start=0, end=1, label="test")]
        expected = {
            "confidence": -1,
            "context_similarity": -1,
            "cui": -1,
            "detected_name": None,
            "end": 1,
            "id": 0,
            "label": "test",
            "link_candidates": None,
            "start": 0
        }

        serialized = _PipeRunnerImpl.serialize_entities(doc)

        self.assertEqual(expected, serialized._.ents[0])

    def test_deserialize_entities(self):
        doc = self.nlp.make_doc(self.text)
        doc._.ents = [{
            "confidence": -1,
            "context_similarity": -1,
            "cui": -1,
            "detected_name": None,
            "end": 1,
            "id": 0,
            "label": "test",
            "link_candidates": None,
            "start": 0
        }]

        deserialized = _PipeRunnerImpl.deserialize_entities(doc)._.ents[0]

        self.assertEqual(-1, deserialized._.confidence)
        self.assertEqual(-1, deserialized._.context_similarity)
        self.assertEqual(-1, deserialized._.cui)
        self.assertEqual(None, deserialized._.detected_name)
        self.assertEqual(1, deserialized.end)
        self.assertEqual(0, deserialized._.id)
        self.assertEqual("test", deserialized.label_)
        self.assertEqual(None, deserialized._.link_candidates)
        self.assertEqual(0, deserialized.start)


class _PipeRunnerImpl(PipeRunner):

    def __call__(self, doc):
        return doc




import pandas
import numpy as np
import datetime
import logging
import re
from typing import Optional, List, Dict

from medcat.pipe import Pipe
from medcat.cdb import CDB
from medcat.config import Config
from medcat.preprocessing.tokenizers import spacy_split_all
from medcat.preprocessing.cleaners import prepare_name
from medcat.preprocessing.taggers import tag_skip_and_punct
from medcat.utils.loggers import add_handlers

PH_REMOVE = re.compile("(\s)\([a-zA-Z]+[^\)\(]*\)($)")


class CDBMaker(object):
    r''' Given a CSV as shown in https://github.com/CogStack/MedCAT/tree/master/examples/<example> it creates a CDB or
    updates an exisitng one.

    Args:
        config (`medcat.config.Config`):
            Global config for MedCAT.
        cdb (`medcat.cdb.CDB`, optional):
            If set the `CDBMaker` will updat the existing `CDB` with new concepts in the CSV.
        name_max_words (`int`, defaults to `20`):
            Names with more words will be skipped during the build of a CDB
    '''
    log = logging.getLogger(__package__)
    log = add_handlers(log)

    def __init__(self, config: Config, cdb: Optional[CDB] = None) -> None:
        self.config = config
        # Set log level
        self.log.setLevel(self.config.general['log_level'])

        # To make life a bit easier
        self.cnf_cm = config.cdb_maker

        if cdb is None:
            self.cdb = CDB(config=self.config)
        else:
            self.cdb = cdb

        # Build the required spacy pipeline
        self.pipe = Pipe(tokenizer=spacy_split_all, config=config)
        self.pipe.add_tagger(tagger=tag_skip_and_punct,
                             name='skip_and_punct',
                             additional_fields=['is_punct'])

    def prepare_csvs(self,
                     csv_paths: List[str],
                     sep: str = ',',
                     encoding: Optional[str] = None,
                     escapechar: Optional[str] = None,
                     index_col: bool = False,
                     full_build: bool = False,
                     only_existing_cuis: bool = False, **kwargs) -> CDB:
        r''' Compile one or multiple CSVs into a CDB.

        Args:
            csv_paths (`List[str]`):
                An array of paths to the csv files that should be processed
            full_build (`bool`, defaults to `True`):
                If False only the core portions of the CDB will be built (the ones required for
                the functioning of MedCAT). If True, everything will be added to the CDB - this
                usually includes concept descriptions, various forms of names etc (take care that
                this option produces a much larger CDB).
            sep (`str`, defaults to `,`):
                If necessary a custom separator for the csv files
            encoding (`str`, optional):
                Encoding to be used for reading the CSV file
            escapechar (`str`, optional):
                Escape char for the CSV
            index_col (`bool`, defaults_to `False`):
                Index column for pandas read_csv
            only_existing_cuis (`bool`, defaults to False):
                If True no new CUIs will be added, but only linked names will be extended. Mainly used when
                enriching names of a CDB (e.g. SNOMED with UMLS terms).
        Return:
            `medcat.cdb.CDB` with the new concepts added.

        Note:
            **kwargs:
                Will be passed to pandas for CSV reading
            csv:
                Examples of the CSV used to make the CDB can be found on [GitHub](link)
        '''

        useful_columns = ['cui', 'name', 'ontologies', 'name_status', 'type_ids', 'description']
        name_status_options = {'A', 'P', 'N'}

        for csv_path in csv_paths:
            # Read CSV, everything is converted to strings
            df = pandas.read_csv(csv_path, sep=sep, encoding=encoding, escapechar=escapechar, index_col=index_col, dtype=str, **kwargs)
            df = df.fillna('')

            # Find which columns to use from the CSV
            cols: List = []
            col2ind = {}
            for col in list(df.columns):
                if str(col).lower().strip() in useful_columns:
                    col2ind[str(col).lower().strip()] = len(cols)
                    cols.append(col)

            self.log.info("Started importing concepts from: {}".format(csv_path))
            _time = None # Used to check speed
            _logging_freq = np.ceil(len(df[cols]) / 100)
            for row_id, row in enumerate(df[cols].values):
                if row_id % _logging_freq == 0:
                    # Print some stats
                    if _time is None:
                        # Add last time if it does not exist
                        _time = datetime.datetime.now()
                    # Get current time
                    ctime = datetime.datetime.now()
                    # Get time difference
                    timediff = ctime - _time
                    self.log.info("Current progress: {:.0f}% at {:.3f}s per {} rows".format(
                        (row_id / len(df)) * 100, timediff.microseconds/10**6 + timediff.seconds, (len(df[cols]) // 100)))
                    # Set previous time to current time
                    _time = ctime

                # This must exist
                cui = row[col2ind['cui']].strip().upper()

                if not only_existing_cuis or (only_existing_cuis and cui in self.cdb.cui2names):
                    if 'ontologies' in col2ind:
                        ontologies = set([ontology.strip() for ontology in row[col2ind['ontologies']].upper().split(self.cnf_cm['multi_separator']) if
                                         len(ontology.strip()) > 0])
                    else:
                        ontologies = set()

                    if 'name_status' in col2ind:
                        name_status = row[col2ind['name_status']].strip().upper()

                        # Must be allowed
                        if name_status not in name_status_options:
                            name_status = 'A'
                    else:
                        # Defaults to A - meaning automatic
                        name_status = 'A'

                    if 'type_ids' in col2ind:
                        type_ids = set([type_id.strip() for type_id in row[col2ind['type_ids']].upper().split(self.cnf_cm['multi_separator']) if
                                        len(type_id.strip()) > 0])
                    else:
                        type_ids = set()

                    # Get the ones that do not need any changing
                    if 'description' in col2ind:
                        description = row[col2ind['description']].strip()
                    else:
                        description = ""

                    # We can have multiple versions of a name
                    names: Dict = {} # {'name': {'tokens': [<str>], 'snames': [<str>]}}

                    raw_names = [raw_name.strip() for raw_name in row[col2ind['name']].split(self.cnf_cm['multi_separator']) if
                                 len(raw_name.strip()) > 0]
                    for raw_name in raw_names:
                        raw_name = raw_name.strip()
                        prepare_name(raw_name, self.pipe.spacy_nlp, names, self.config)

                        if self.config.cdb_maker.get('remove_parenthesis', 0) > 0 and name_status == 'P':
                            # Should we remove the content in parenthesis from primary names and add them also
                            raw_name = PH_REMOVE.sub(" ", raw_name).strip()
                            if len(raw_name) >= self.config.cdb_maker['remove_parenthesis']:
                                prepare_name(raw_name, self.pipe.spacy_nlp, names, self.config)

                    self.cdb.add_concept(cui=cui, names=names, ontologies=ontologies, name_status=name_status, type_ids=type_ids,
                                         description=description, full_build=full_build)
                    # DEBUG
                    self.log.debug("\n\n**** Added\n CUI: %s\n Names: %s\n Ontologies: %s\n Name status: %s\n Type IDs: %s\n Description: %s\n Is full build: %s",
                                   cui, names, ontologies, name_status, type_ids, description, full_build)

        return self.cdb

    def destroy_pipe(self) -> None:
        self.pipe.destroy()



from py2neo import Graph
import getpass
from collections import defaultdict


class NeoConnector:
    def __init__(self, uri, user, password=None):
        if password is None:
            password = getpass.getpass("Password:")
        self.graph = Graph(uri, auth=(user, password))

    def execute(self, query):
        r = self.graph.run(query)
        return r

    def bucket_concepts(self, data, bucket_size_seconds):
        entities = data['entities']

        _bucket = []
        _concepts = set()
        start_time = -1
        new_stream = []
        # Sort entities
        entities.sort(key=lambda ent: ent['timestamp'])
        for ent in entities:
            if start_time == -1:
                start_time = ent['timestamp']

            if ent['timestamp'] - start_time >= bucket_size_seconds:
                # Add to stream
                new_stream.extend(_bucket)
                _bucket = []
                _concepts = set()
                start_time = ent['timestamp']

                t_ent = dict(new_stream[-1])
                t_ent['timestamp'] += 1
                t_ent['name'] = '<SEP>'
                t_ent['conceptId'] = '<SEP>'
                new_stream.append(t_ent)

            if ent['conceptId'] not in _concepts:
                _bucket.append(ent)
                _concepts.add(ent['conceptId'])

        if _bucket:
            new_stream.extend(_bucket)

        data['entities'] = new_stream

    def get_all_patients(self, concepts, limit=1000, require_time=False, ignore_meta=False):
        r''' Return all patients having all concepts

        Args:
            concepts
            limit
            require_time:
                If set only concepts that have the timestamp property will be used.
        '''

        q = "WITH [{}] AS cs ".format(",".join(["'{}'".format(c) for c in concepts]))
        if not require_time:
            q += '''MATCH (c:Concept)<-[:HAS '''
            if not ignore_meta:
                q += '''{metaPresence: 'True', metaSubject: 'Patient'}'''
            q += ''']-(:Document)<-[:HAS]-(pt:Patient)
            WHERE c.conceptId in cs
            WITH pt, size(cs) as inputCnt, count(DISTINCT c) as cnt
            WHERE cnt = inputCnt
            '''
        else:
            q += '''MATCH (c:Concept)<-[r:HAS {metaPresence: 'True', metaSubject:
            'Patient'}]-(:Document)<-[:HAS]-(pt:Patient) \n
            WHERE c.conceptId in cs AND exists(r.timestamp) \n
            WITH pt, size(cs) as inputCnt, count(DISTINCT c) as cnt \n
            WHERE cnt = inputCnt \n
            '''

        q += ' RETURN pt LIMIT {}'.format(limit)
        data = self.execute(q).data() # Do not like this too much 

        return [n['pt']['patientId'] for n in data], q

    def get_all_concepts_from(self, patient_id=None, document_id=None,
            limit=1000, bucket_size_seconds=None, min_count=0, meta_requirements=None, require_time=True):
        r''' Returns all concepts belonging to a document or patient
        given the concept type (if none all are retruned).
        '''

        if patient_id is not None:
            q = 'MATCH (patient:Patient {patientId: "%s"})-[:HAS]->' % patient_id \
                + '(document:Document)-[has:HAS]->(concept:Concept) \n'
        elif document_id is not None:
            q = 'MATCH (patient:Patient)-[:HAS]->(document:Document {documentId: "%s"})' % document_id \
                + '-[has:HAS]->(concept:Concept) \n'
        else:
            raise Exception("patient_id or document_id are required")
        q += 'RETURN patient, document, concept, has LIMIT %s \n' % limit

        data = self.execute(q).data() # Do not like this too much 
        out = None
        if len(data) > 0:
            out = {'patient': dict(data[0]['patient']),
                   'entities': []}

            cnt = defaultdict(int)
            for row in data:
                if meta_requirements is None or \
                   all([row['has'][meta] == value for meta,value in meta_requirements.items()]):
                    if not require_time or 'timestamp' in row['has']:
                        ent = dict(row['concept']) # Take everything from concept
                        ent['documentId'] = row['document']['documentId']
                        ent.update(row['has']) # add all the stuff from the meta ann

                        out['entities'].append(ent)
                        cnt[ent['conceptId']] += 1

            # Cleanup based on min_count
            new_ents = []
            for ent in out['entities']:
                if cnt[ent['conceptId']] >= min_count:
                    ent['count'] = cnt[ent['conceptId']]
                    new_ents.append(ent)
            out['entities'] = new_ents

            if bucket_size_seconds is not None:
                self.bucket_concepts(data=out, bucket_size_seconds=bucket_size_seconds)

        return out, q

    def get_all_patients_descend(self, concepts, limit=1000, require_time=False):
        r''' Return all patients having all descendant concepts under the ancestor concept

        Args:
            concepts - ancestor top-level concepts
            limit
            require_time:
                If set only concepts that have the timestamp property will be used.
        Output:
            lists of patients with attached SNOMED concepts
        '''

        q = "WITH [{}] AS ancestor ".format(",".join(["'{}'".format(c) for c in concepts]))
        if not require_time:
            q += '''MATCH (n:Concept)-[:IS_A*0..5]->(m:Concept)
                    WHERE m.conceptId IN ancestor ## get the ancestor and the children
                    WITH [n.conceptId] AS lineage ## pass the lineage to patient match
                    MATCH (c:Concept)<-[r:HAS {metaPresence: 'True', metaSubject: 'Patient'}]-(d:Document)<-[q:HAS]-(pt:Patient)
                    WHERE c.conceptId in lineage    
                    '''
        else:
            q += '''MATCH (n:Concept)-[:IS_A*0..5]->(m:Concept)
                    WHERE m.conceptId IN ancestor ## get the ancestor and the children
                    WITH [n.conceptId] AS lineage ## pass the lineage to patient match
                    MATCH (c:Concept)<-[r:HAS {metaPresence: 'True', metaSubject: 'Patient'}]-(d:Document)<-[q:HAS]-(pt:Patient)
                    WHERE c.conceptId in lineage AND exists(r.timestamp)
                    '''

        q += ' RETURN pt.patientId, pt.sex, c.conceptId, c.name, r.timestamp LIMIT {}'.format(limit)
        data = self.execute(q).data() # Do not like this too much 

        return [n['pt']['patientId'] for n in data], q



import warnings
import functools
from typing import Callable


def deprecated(message: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapped(*args, **kwargs) -> Callable:
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn("Function {} has been deprecated.{}".format(func.__name__, " " + message if message else ""))
            warnings.simplefilter("default", DeprecationWarning)
            return func(*args, **kwargs)
        return wrapped
    return decorator


def check_positive(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapped(*args, **kwargs) -> Callable:
        for idx, arg in enumerate(args):
            if isinstance(arg, int):
                if arg < 1:
                    raise ValueError("Argument at position %s is not a positive integer" % idx)
        for key, value in kwargs.items():
            if isinstance(value, int) and value < 1:
                raise ValueError("Argument '%s' is not a positive integer" % key)
        return func(*args, **kwargs)
    return wrapped



import os
import unittest
import tempfile
import json
from unittest.mock import patch
from tests.helper import AsyncMock
from medcat.utils.checkpoint import Checkpoint, CheckpointConfig, CheckpointManager
from medcat.cdb import CDB


class CheckpointTest(unittest.TestCase):

    def test_from_latest(self):
        dir_path = os.path.join(os.path.dirname(__file__), "..", "resources", "checkpoints", "cat_train", "1643822916")

        checkpoint = Checkpoint.from_latest(dir_path)

        self.assertEqual(2, checkpoint.steps)
        self.assertEqual(20, checkpoint.count)
        self.assertEqual(1, checkpoint.max_to_keep)

    @patch("medcat.cdb.CDB.load", return_value="cdb_object")
    def test_restore_latest_cdb(self, cdb_load):
        dir_path = os.path.join(os.path.dirname(__file__), "..", "resources", "checkpoints", "cat_train", "1643822916")
        checkpoint = Checkpoint(dir_path=dir_path, steps=1, max_to_keep=5)

        cdb = checkpoint.restore_latest_cdb()

        self.assertEqual("cdb_object", cdb)
        self.assertEqual(1, checkpoint.steps)
        self.assertEqual(20, checkpoint.count)
        self.assertEqual(5, checkpoint.max_to_keep)
        cdb_load.assert_called_with(os.path.abspath(os.path.join(dir_path, "checkpoint-2-20")))

    @patch("medcat.cdb.CDB")
    def test_save(self, cdb):
        dir_path = tempfile.TemporaryDirectory()
        checkpoint = Checkpoint(dir_path=dir_path.name, steps=1, max_to_keep=1)

        checkpoint.save(cdb, 1)

        cdb.save.assert_called()
        self.assertEqual(1, checkpoint.steps)
        self.assertEqual(1, checkpoint.max_to_keep)
        self.assertEqual(1, checkpoint.count)

    def test_validation_on_steps(self):
        with self.assertRaises(Exception) as e1:
            Checkpoint(dir_path="dir_path", steps=0, max_to_keep=1)
        self.assertEqual("Argument 'steps' is not a positive integer", str(e1.exception))

        with self.assertRaises(Exception) as e2:
            checkpoint = Checkpoint(dir_path="dir_path", steps=1000, max_to_keep=1)
            checkpoint.steps = 0
        self.assertEqual("Argument at position 0 is not a positive integer", str(e2.exception))

    def test_validation_on_max_to_keep(self):
        with self.assertRaises(Exception) as e1:
            Checkpoint(dir_path="dir_path", steps=1000, max_to_keep=-1)
        self.assertEqual("Argument 'max_to_keep' is not a positive integer", str(e1.exception))

        with self.assertRaises(Exception) as e2:
            checkpoint = Checkpoint(dir_path="dir_path", steps=1000, max_to_keep=1)
            checkpoint.max_to_keep = -1
        self.assertEqual("Argument at position 0 is not a positive integer", str(e2.exception))

    def test_create_checkpoint(self):
        ckpt_out_dir_path = tempfile.TemporaryDirectory().name
        ckpt_config = {
            'output_dir': ckpt_out_dir_path,
            'steps': 1000,
            "max_to_keep": 5,
        }

        checkpoint = CheckpointManager("cat_train", CheckpointConfig(**ckpt_config)).create_checkpoint()

        self.assertTrue("cat_train" in checkpoint.dir_path)
        self.assertEqual(1000, checkpoint.steps)
        self.assertEqual(0, checkpoint.count)
        self.assertEqual(5, checkpoint.max_to_keep)

    def test_get_latest_checkpoint(self):
        ckpt_out_dir_path = os.path.join(os.path.dirname(__file__), "..", "resources", "checkpoints", "cat_train_supervised")
        ckpt_config = {
            'output_dir': ckpt_out_dir_path,
            'steps': 1000,
            "max_to_keep": 5,
        }

        checkpoint = CheckpointManager("cat_train_supervised", CheckpointConfig(**ckpt_config)).get_latest_checkpoint(ckpt_out_dir_path)

        self.assertTrue("1643823460" in checkpoint.dir_path)
        self.assertEqual(1, checkpoint.count)
        self.assertEqual(1000, checkpoint.steps)
        self.assertEqual(5, checkpoint.max_to_keep)

    def test_get_latest_training_dir(self):
        ckpt_out_dir_path = os.path.join(os.path.dirname(__file__), "..", "resources", "checkpoints", "cat_train")

        ckpt_dir_path = CheckpointManager.get_latest_training_dir(ckpt_out_dir_path)

        self.assertTrue("1643823460" in ckpt_dir_path)



from medcat.datasets.helpers import encode_examples
from datasets import load_dataset
from transformers.data.data_collator import DataCollatorWithPadding
from transformers import AutoTokenizer
from torch.utils.data import DataLoader
from medcat.meta_cat import MetaCAT
import torch

CACHE_DIR = './'
DATASETS_CLASS_PATH = "../medcat/datasets/medcat_annotations.py"
ANNOTATIONS_PATH = ''
TOKENIZER_PATH = ''
MAX_SEQ_LEN = 100
BATCH_SIZE = 1024
DEVICE = torch.device('cuda')
META_CAT_SAVE_DIR = ''

tokenizer = AutoTokenizer.from_pretrained(TOKENIZER_PATH)

dataset = load_dataset(DATASETS_CLASS_PATH, data_files=ANNOTATIONS_PATH, name='pickle', split='train', cache_dir=CACHE_DIR)

encoded_dataset = dataset.map(
        lambda examples: encode_examples(examples['context_left'], examples['context_center'], examples['context_right'], tokenizer, MAX_SEQ_LEN),
        batched=True,
        remove_columns=['context_center', 'context_left', 'context_right', 'document_id', 'id'])

data_collator = DataCollatorWithPadding(tokenizer=tokenizer, max_length=MAX_SEQ_LEN)

dataset_loader = DataLoader(encoded_dataset, batch_size=BATCH_SIZE, shuffle=False, collate_fn=data_collator)

meta_cat = MetaCAT(save_dir=META_CAT_SAVE_DIR, tokenizer=tokenizer)
meta_cat.load()
meta_cat.model.eval()
meta_cat.model.to(DEVICE)

logits = []
for i, batch in enumerate(dataset_loader):
    batch = {k:v.to(DEVICE) for k,v in batch.items()}
    out = meta_cat.model(**batch)
    logits.extend(out.to('cpu').detach().numpy().tolist())

    if i % 1000 == 0:
        print("Done with batch: {}".format(i))

predictions = np.argmax(logits, axis=1)



import torch
import numpy as np


def deid_document(text, tokenizer, model, cdb, verbose=False):
    label2name = {v:k for k,v in tokenizer.label_map.items()}

    tkns_raw = tokenizer.hf_tokenizer(text, return_offsets_mapping=True, add_special_tokens=False)
    offsets = tkns_raw['offset_mapping']
    del tkns_raw['offset_mapping']
    preds = [0] * len(offsets)
    for i in range(((len(tkns_raw['input_ids']) - 1) // 500) + 1):
        tkns = {k:torch.tensor([v[i*500:(i+1)*500]]).to(model.device) for k,v in tkns_raw.items()}
        _preds = model(**tkns).logits[0].detach().cpu().numpy()
        _preds = np.argmax(_preds, axis=1)
        _preds[_preds == 1] = 0 # Ignore the token 1
        preds[i*500:i*500 + len(_preds)] = list(_preds)

    new_text = ""
    start = 0
    end = -1
    p_end = 0
    c_ent = False
    for i in range(len(offsets)):
        if preds[i] != 0:
            if not c_ent:
                start = offsets[i][0]

            # Find index of next token that is not subtoken
            n_real = i
            for k in range(i+1, len(offsets)):
                if tokenizer.id2type[tkns_raw['input_ids'][k]] == 'start':
                    n_real == k
                    break

            if i == len(offsets) -1 or preds[i] != preds[i+1]:
                end = offsets[i][1]
                for j in range(i+1, len(offsets)):
                    if tokenizer.id2type[tkns_raw['input_ids'][j]] == 'start':
                        end = offsets[j-1][1]
                        break
                if verbose:
                    deid = "***{} -> {}***".format(text[start:end], cdb.get_name(label2name[preds[i]]))
                else:
                    deid = "***{}***".format(cdb.get_name(label2name[preds[i]]))
                new_text = new_text + text[p_end:start] + deid
                p_end = end
                c_ent = False
            else:
                c_ent = True
    new_text = new_text + text[p_end:]

    return new_text



import re
import os
import spacy
from typing import Any, List, Dict, cast, Iterable, Union, Pattern
from spacy.tokenizer import Tokenizer
from spacy.language import Language
from spacy.tokens import Doc
from tokenizers import ByteLevelBPETokenizer
from transformers.models.bert.tokenization_bert_fast import BertTokenizerFast
from medcat.config import Config


def spacy_extended(nlp: Language) -> Tokenizer:
    infix_re_list = ('\\.\\.+',
    '''(?<=[A-Za-z]{1})[\-_;\,\/~]+(?=[A-Za-z]{1})|(?<=[0-9]{1})[\-_;\,\/]+(?=[A-Za-z]{1})|(?<=[A-Za-z]{1})[\-_;\,\/]+(?=[0-9]{1})|\d{2,4}[\-\s_\*]\d{1,2}[\-\s_\*]\d{1,2}|\d{1,2}:\d{1,2}:\d{1,2}|\d{1,2}:\d{2}'''
     '…',
     '[\\p{So}]',
     '(?<=[[[\\p{Ll}&&\\p{Latin}]||[ёа-я]||[әөүҗңһ]||[α-ωάέίόώήύ]||[\\p{L}&&\\p{Bengali}]||[\\p{L}&&\\p{Hebrew}]||[\\p{L}&&\\p{Arabic}]||[\\p{L}&&\\p{Sinhala}]]])\\.(?=[[[\\p{Lu}&&\\p{Latin}]||[ЁА-Я]||[ӘӨҮҖҢҺ]||[Α-ΩΆΈΊΌΏΉΎ]||[\\p{L}&&\\p{Bengali}]||[\\p{L}&&\\p{Hebrew}]||[\\p{L}&&\\p{Arabic}]||[\\p{L}&&\\p{Sinhala}]]])',
     '(?<=[[[\\p{Lu}&&\\p{Latin}]||[ЁА-Я]||[ӘӨҮҖҢҺ]||[Α-ΩΆΈΊΌΏΉΎ]||[\\p{Ll}&&\\p{Latin}]||[ёа-я]||[әөүҗңһ]||[α-ωάέίόώήύ]||[\\p{L}&&\\p{Bengali}]||[\\p{L}&&\\p{Hebrew}]||[\\p{L}&&\\p{Arabic}]||[\\p{L}&&\\p{Sinhala}]]]),(?=[[[\\p{Lu}&&\\p{Latin}]||[ЁА-Я]||[ӘӨҮҖҢҺ]||[Α-ΩΆΈΊΌΏΉΎ]||[\\p{Ll}&&\\p{Latin}]||[ёа-я]||[әөүҗңһ]||[α-ωάέίόώήύ]||[\\p{L}&&\\p{Bengali}]||[\\p{L}&&\\p{Hebrew}]||[\\p{L}&&\\p{Arabic}]||[\\p{L}&&\\p{Sinhala}]]])',
     '(?<=[[[\\p{Lu}&&\\p{Latin}]||[ЁА-Я]||[ӘӨҮҖҢҺ]||[Α-ΩΆΈΊΌΏΉΎ]||[\\p{Ll}&&\\p{Latin}]||[ёа-я]||[әөүҗңһ]||[α-ωάέίόώήύ]||[\\p{L}&&\\p{Bengali}]||[\\p{L}&&\\p{Hebrew}]||[\\p{L}&&\\p{Arabic}]||[\\p{L}&&\\p{Sinhala}]]])[?";:=,.]*(?:-|–|—|--|---|——|~)(?=[[[\\p{Lu}&&\\p{Latin}]||[ЁА-Я]||[ӘӨҮҖҢҺ]||[Α-ΩΆΈΊΌΏΉΎ]||[\\p{Ll}&&\\p{Latin}]||[ёа-я]||[әөүҗңһ]||[α-ωάέίόώήύ]||[\\p{L}&&\\p{Bengali}]||[\\p{L}&&\\p{Hebrew}]||[\\p{L}&&\\p{Arabic}]||[\\p{L}&&\\p{Sinhala}]]])',
     '(?<=[[[\\p{Lu}&&\\p{Latin}]||[ЁА-Я]||[ӘӨҮҖҢҺ]||[Α-ΩΆΈΊΌΏΉΎ]||[\\p{Ll}&&\\p{Latin}]||[ёа-я]||[әөүҗңһ]||[α-ωάέίόώήύ]||[\\p{L}&&\\p{Bengali}]||[\\p{L}&&\\p{Hebrew}]||[\\p{L}&&\\p{Arabic}]||[\\p{L}&&\\p{Sinhala}]]"])[:<>=/](?=[[[\\p{Lu}&&\\p{Latin}]||[ЁА-Я]||[ӘӨҮҖҢҺ]||[Α-ΩΆΈΊΌΏΉΎ]||[\\p{Ll}&&\\p{Latin}]||[ёа-я]||[әөүҗңһ]||[α-ωάέίόώήύ]||[\\p{L}&&\\p{Bengali}]||[\\p{L}&&\\p{Hebrew}]||[\\p{L}&&\\p{Arabic}]||[\\p{L}&&\\p{Sinhala}]]])')

    prefix_iter = cast(Iterable[Union[str, Pattern[Any]]], Language.Defaults.prefixes)
    suffix_iter = cast(Iterable[Union[str, Pattern[Any]]], Language.Defaults.suffixes)
    prefix_re = spacy.util.compile_prefix_regex(prefix_iter)
    suffix_re = spacy.util.compile_suffix_regex(suffix_iter)
    infix_re = spacy.util.compile_infix_regex(infix_re_list)


    return Tokenizer(nlp.vocab,
            rules={},
            prefix_search=prefix_re.search,
            suffix_search=suffix_re.search,
            infix_finditer=infix_re.finditer
            )


def spacy_split_all(nlp: Language, config: Config) -> Tokenizer:

    token_characters = r'[^A-Za-z0-9\@]'

    if config.general['diacritics']:
        token_characters = r'[^A-Za-zÀ-ÖØ-öø-ÿ0-9\@]'

    infix_re = re.compile(token_characters)
    suffix_re = re.compile(token_characters + r'$')
    prefix_re = re.compile(r'^' + token_characters)
    return Tokenizer(nlp.vocab,
            rules={},
            token_match=None,
            prefix_search=prefix_re.search,
            suffix_search=suffix_re.search,
            infix_finditer=infix_re.finditer
            )


class WordpieceTokenizer(object):
    """Runs WordPiece tokenziation."""

    def __init__(self, vocab: Any, unk_token: str = "[UNK]", max_input_chars_per_word: int = 200) -> None:
        self.vocab = vocab
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word

    def tokenize(self, text: str) -> List:
        """Tokenizes a piece of text into its word pieces.
        This uses a greedy longest-match-first algorithm to perform tokenization
        using the given vocabulary.
        For example:
          input = "unaffable"
          output = ["un", "##aff", "##able"]
        Args:
          text: A single token or whitespace separated tokens. This should have
            already been passed through `BasicTokenizer.
        Returns:
          A list of wordpiece tokens.
        """

        # Why is convert_to_unicode undefined?
        text = convert_to_unicode(text) # type: ignore # noqa

        output_tokens = []

        # Why is whitespace_tokenize undefined?
        for token in whitespace_tokenize(text): # type: ignore # noqa
            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            is_bad = False
            start = 0
            sub_tokens = []
            while start < len(chars):
                end = len(chars)
                cur_substr = None
                while start < end:
                    substr = "".join(chars[start:end])
                    if start > 0:
                        substr = "##" + substr
                    if substr in self.vocab:
                        cur_substr = substr
                        break
                    end -= 1
                if cur_substr is None:
                    is_bad = True
                    break
                sub_tokens.append(cur_substr)
                start = end

            if is_bad:
                output_tokens.append(self.unk_token)
            else:
                output_tokens.extend(sub_tokens)
        return output_tokens


class SpacyHFTok(object):

    def __init__(self, w2v: Any) -> None:
        self.nlp = spacy.load('en_core_sci_md', disable=['ner', 'parser'])
        self.emb_map = {}
        self.embs: List = []
        for key in w2v.wv.key_to_index.keys():
            self.emb_map[key] = len(self.embs)
            self.embs.append(w2v[key])

        # Add pad
        self.embs.append([0.0] * 300)

    def encode(self, text: str) -> 'SpacyHFDoc':
        doc = self.nlp(text)
        return SpacyHFDoc(doc)

    def token_to_id(self, tok: Any) -> Any:
        return self.emb_map.get(tok, len(self.emb_map) - 1)


class SpacyHFDoc(object):
    def __init__(self, doc: Doc) -> None:
        self.doc = doc
        self.tokens = [x.text for x in self.doc]
        self.offsets = [(x.idx, x.idx+len(x.text)) for x in self.doc]


class TokenizerWrapperBPE(object):
    '''
    '''

    def __init__(self, hf_tokenizers: Any) -> None:
        self.hf_tokenizers = hf_tokenizers

    def __call__(self, text: str) -> Dict:
        res = self.hf_tokenizers.encode(text)

        return {'offset_mapping': res.offsets,
                'input_ids': res.ids,
                'tokens': res.tokens,
                }

    def save(self, dir_path, name='bbpe'):
        self.hf_tokenizers.save_model(dir_path, prefix=name)

    @classmethod
    def load(cls, dir_path, name='bbpe', **kwargs):
        tokenizer = cls()
        vocab_file = os.path.join(dir_path, f'{name}-vocab.json')
        merges_file = os.path.join(dir_path, f'{name}-merges.txt')
        tokenizer.hf_tokenizers = ByteLevelBPETokenizer.from_file(vocab_filename=vocab_file,
                                                                  merges_filename=merges_file,
                                                                  **kwargs)

        return tokenizer


class TokenizerWrapperBERT(object):
    '''
    '''

    def __init__(self, hf_tokenizers=None):
        self.hf_tokenizers = hf_tokenizers

    def __call__(self, text: str) -> Dict:
        res = self.hf_tokenizers.encode_plus(text,
                return_offsets_mapping=True, add_special_tokens=False)

        return {'offset_mapping': res['offset_mapping'],
                'input_ids': res['input_ids'],
                'tokens':  self.hf_tokenizers.convert_ids_to_tokens(res['input_ids']),
                }

    def save(self, dir_path: str, name: str='bert') -> None:
        path = os.path.join(dir_path, name)
        self.hf_tokenizers.save_pretrained(path)

    @classmethod
    def load(cls, dir_path: str, name: str = 'bert', **kwargs) -> Any:
        tokenizer = cls()
        path = os.path.join(dir_path, name)
        tokenizer.hf_tokenizers = BertTokenizerFast.from_pretrained(path, **kwargs)

        return tokenizer
